#!/usr/bin/python3
#
# Copyright (C) 2020-2022 Canonical, Ltd.
# Author: Łukasz 'sil2100' Zemczak <lukasz.zemczak@canonical.com>
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import subprocess
import typing
from typing import Dict, List, Optional, Set

from . import utils
from ..configmanager import ConfigurationError
import netplan


# PCIDevice class originates from mlnx_switchdev_mode/sriovify.py
# Copyright 2019 Canonical Ltd, Apache License, Version 2.0
# https://github.com/openstack-charmers/mlnx-switchdev-mode
class PCIDevice(object):
    """Helper class for interaction with a PCI device"""

    def __init__(self, pci_addr: str):
        """Initialise a new PCI device handler
        :param pci_addr: PCI address of device
        :type: str
        """
        self.pci_addr = pci_addr

    @property
    def sys(self) -> str:
        """sysfs path (can be overridden for testing)
        :return: full path to /sys filesystem
        :rtype: str
        """
        return "/sys"

    @property
    def path(self) -> str:
        """/sys path for PCI device
        :return: full path to PCI device in /sys filesystem
        :rtype: str
        """
        return os.path.join(self.sys, "bus/pci/devices", self.pci_addr)

    def subpath(self, subpath: str) -> str:
        """/sys subpath helper for PCI device
        :param subpath: subpath to construct path for
        :type: str
        :return: self.path + subpath
        :rtype: str
        """
        return os.path.join(self.path, subpath)

    @property
    def driver(self) -> str:
        """Kernel driver for PCI device
        :return: kernel driver in use for device
        :rtype: str
        """
        driver = ''
        if os.path.exists(self.subpath("driver")):
            driver = os.path.basename(os.readlink(self.subpath("driver")))
        return driver

    @property
    def bound(self) -> bool:
        """Determine if device is bound to a kernel driver
        :return: whether device is bound to a kernel driver
        :rtype: bool
        """
        return os.path.exists(self.subpath("driver"))

    @property
    def is_pf(self) -> bool:
        """Determine if device is a SR-IOV Physical Function
        :return: whether device is a PF
        :rtype: bool
        """
        return os.path.exists(self.subpath("sriov_numvfs"))

    @property
    def is_vf(self) -> bool:
        """Determine if device is a SR-IOV Virtual Function
        :return: whether device is a VF
        :rtype: bool
        """
        return os.path.exists(self.subpath("physfn"))

    @property
    def vf_addrs(self) -> list:
        """List Virtual Function addresses associated with a Physical Function
        :return: List of PCI addresses of Virtual Functions
        :rtype: list[str]
        """
        vf_addrs = []
        i = 0
        while True:
            try:
                vf_addrs.append(
                    os.path.basename(
                        os.readlink(self.subpath("virtfn{}".format(i)))
                    )
                )
            except FileNotFoundError:
                break
            i += 1
        return vf_addrs

    @property
    def vfs(self) -> list:
        """List Virtual Function associated with a Physical Function
        :return: List of PCI devices of Virtual Functions
        :rtype: list[PCIDevice]
        """
        return [PCIDevice(addr) for addr in self.vf_addrs]

    def devlink_set(self, obj_name: str, prop: str, value: str):
        """Set devlink options for the PCI device
        :param obj_name: devlink object to set options on
        :type: str
        :param prop: property to set
        :type: str
        :param value: value to set for property
        :type: str
        """
        subprocess.check_call(
            [
                "/sbin/devlink",
                "dev",
                obj_name,
                "set",
                "pci/{}".format(self.pci_addr),
                prop,
                value,
            ]
        )

    def devlink_eswitch_mode(self) -> str:
        """Query eswitch mode via devlink for the PCI device
        :return: the eswitch mode or '__undetermined' if it can't be retrieved
        :rtype: str
        """
        pci = f"pci/{self.pci_addr}"
        try:
            output = subprocess.check_output(
                [
                    "/sbin/devlink",
                    "-j",
                    "dev",
                    "eswitch",
                    "show",
                    pci,
                ],
                stderr=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError:
            return '__undetermined'

        json_output = json.loads(output)

        # The JSON document looks like this when the 'mode' is available:
        # {"dev":{"pci/0000:03:00.0":{"mode":"switchdev"}}}
        # and like this when it's not available
        # {"dev":{}}
        return json_output.get("dev", {}).get(pci, {}).get('mode', '__undetermined')

    def __str__(self) -> str:
        """String represenation of object
        :return: PCI address of string
        :rtype: str
        """
        return self.pci_addr


def bind_vfs(vfs: typing.Iterable[PCIDevice], driver):
    """Bind unbound VFs to driver."""
    bound_vfs = []
    for vf in vfs:
        if not vf.bound:
            with open("/sys/bus/pci/drivers/{}/bind".format(driver), "wt") as f:
                f.write(vf.pci_addr)
                bound_vfs.append(vf)
    return bound_vfs


def unbind_vfs(vfs: typing.Iterable[PCIDevice], driver) -> typing.Iterable[PCIDevice]:
    """Unbind bound VFs from driver."""
    unbound_vfs = []
    for vf in vfs:
        if vf.bound:
            with open("/sys/bus/pci/drivers/{}/unbind".format(driver), "wt") as f:
                f.write(vf.pci_addr)
                unbound_vfs.append(vf)
    return unbound_vfs


def _interface_matches(netdef: netplan.NetDefinition, interface: str) -> bool:
    return netdef._match_interface(
        iface_name=interface,
        iface_driver=utils.get_interface_driver_name(interface),
        iface_mac=utils.get_interface_macaddress(interface))


def _get_interface_name_for_netdef(netdef: netplan.NetDefinition) -> Optional[str]:
    """
    Try to match a netdef with the real system network interface.
    Throws ConfigurationError if there is more than one match.
    """
    interfaces: List[str] = utils.get_interfaces()
    if netdef._has_match:
        # now here it's a bit tricky
        set_name: str = netdef.set_name
        if set_name and set_name in interfaces:
            # if we had a match: stanza and set-name: this means we should
            # assume that, if found, the interface has already been
            # renamed - use the new name
            return set_name
        else:
            matches: Set[str] = set()
            # we walk through all the system interfaces to determine if there is
            # more than one matched interface
            for interface in interfaces:
                if not _interface_matches(netdef, interface):
                    continue
                # we have a matching PF
                # error out if we matched more than one
                if len(matches) > 1:
                    raise ConfigurationError('matched more than one interface for a PF device: %s' % netdef.id)
                matches.add(interface)
            if matches:
                return list(matches)[0]
    else:
        # no match field, assume entry name is the interface name
        if netdef.id in interfaces:
            return netdef.id

    return None


def _get_pci_slot_name(netdev):
    """
    Read PCI slot name for given interface name
    """
    uevent_path = os.path.join('/sys/class/net', netdev, 'device/uevent')
    try:
        with open(uevent_path) as f:
            pci_slot_name = None
            for line in f.readlines():
                line = line.strip()
                if line.startswith('PCI_SLOT_NAME='):
                    pci_slot_name = line.split('=', 2)[1]
                    return pci_slot_name
    except IOError as e:
        raise RuntimeError('failed parsing PCI slot name for %s: %s' % (netdev, str(e)))


def _get_physical_functions(np_state: netplan.State) -> Dict[str, str]:
    """
    Go through the list of netplan ethernet devices and identify which are
    PFs matching them with actual network interfaces.
    """
    pfs = {}
    for netdef in np_state.ethernets.values():
        # If the sriov_link is present, the interface is a VF and link is the PF
        if link := netdef.links.get('sriov'):
            if iface := _get_interface_name_for_netdef(np_state[link.id]):
                pfs[link.id] = iface
        else:
            # If a netdef also defines the embedded_switch_mode key we consider it's a PF
            # This enables us to change the eswitch mode even when the PF has no VFs.
            if netdef._embedded_switch_mode:
                if iface := _get_interface_name_for_netdef(netdef):
                    pfs[netdef.id] = iface

            # If the netdef has any (positive) number of VFs that's because it's a PF
            try:
                count = netdef._vf_count
            except netplan.NetplanException as e:
                raise ConfigurationError(str(e))
            if count > 0:
                if iface := _get_interface_name_for_netdef(netdef):
                    pfs[netdef.id] = iface

    return pfs


def _get_vf_number_per_pf(np_state: netplan.State) -> Dict[str, int]:
    """
    Go through the list of netplan ethernet devices and identify which ones
    have VFs. netdef._vf_count ultimately calls _netplan_state_get_vf_count_for_def
    from libnetplan which return MAX(sriov_explicit_vf_count, number of VF netdefs).
    """
    vf_counts = {}
    for netdef in np_state.ethernets.values():
        try:
            count = netdef._vf_count
        except netplan.NetplanException as e:
            raise ConfigurationError(str(e))
        if count > 0:
            if iface := _get_interface_name_for_netdef(netdef):
                vf_counts[iface] = count

    return vf_counts


def _get_virtual_functions(np_state: netplan.State) -> Set[str]:
    """
    Go through the list of netplan ethernet devices and identify which ones
    are virtual functions
    """
    vfs = set()
    for netdef in np_state.ethernets.values():
        # If the sriov_link is present and the PF is also present in the system we save the VF
        if link := netdef.links.get('sriov'):
            if _get_interface_name_for_netdef(np_state[link.id]):
                vfs.add(netdef.id)
    return vfs


def set_numvfs_for_pf(pf, vf_count):
    """
    Allocate the required number of VFs for the selected PF.
    """
    if vf_count > 256:
        raise ConfigurationError(
            'cannot allocate more VFs for PF %s than the SR-IOV maximum: %s > 256' % (pf, vf_count))

    devdir = os.path.join('/sys/class/net', pf, 'device')
    numvfs_path = os.path.join(devdir, 'sriov_numvfs')
    totalvfs_path = os.path.join(devdir, 'sriov_totalvfs')
    try:
        with open(totalvfs_path) as f:
            vf_max = int(f.read().strip())
    except IOError as e:
        raise RuntimeError('failed parsing sriov_totalvfs for %s: %s' % (pf, str(e)))
    except ValueError:
        raise RuntimeError('invalid sriov_totalvfs value for %s' % pf)

    if vf_count > vf_max:
        raise ConfigurationError(
            'cannot allocate more VFs for PF %s than supported: %s > %s (sriov_totalvfs)' % (pf, vf_count, vf_max))

    try:
        with open(numvfs_path, 'w') as f:
            f.write(str(vf_count))
    except IOError as e:
        bail = True
        if e.errno == 16:  # device or resource busy
            logging.warning('device or resource busy while setting sriov_numvfs for %s, trying workaround' % pf)
            try:
                # doing this in two open/close sequences so that
                # it's as close to writing via shell as possible
                with open(numvfs_path, 'w') as f:
                    f.write('0')
                with open(numvfs_path, 'w') as f:
                    f.write(str(vf_count))
            except IOError as e_inner:
                e = e_inner
            else:
                bail = False
        if bail:
            raise RuntimeError('failed setting sriov_numvfs to %s for %s: %s' % (vf_count, pf, str(e)))

    return True


def perform_hardware_specific_quirks(pf):
    """
    Perform any hardware-specific quirks for the given SR-IOV device to make
    sure all the VF-count changes are applied.
    """
    devdir = os.path.join('/sys/class/net', pf, 'device')
    try:
        with open(os.path.join(devdir, 'vendor')) as f:
            device_id = f.read().strip()[2:]
        with open(os.path.join(devdir, 'device')) as f:
            vendor_id = f.read().strip()[2:]
    except IOError as e:
        raise RuntimeError('could not determine vendor and device ID of %s: %s' % (pf, str(e)))

    combined_id = ':'.join([vendor_id, device_id])
    quirk_devices = ()  # TODO: add entries to the list
    if combined_id in quirk_devices:  # pragma: nocover (empty quirk_devices)
        # some devices need special handling, so this is the place

        # Currently this part is empty, but has been added as a preemptive
        # measure, as apparently a lot of SR-IOV cards have issues with
        # dynamically allocating VFs. Some cards seem to require a full
        # kernel module reload cycle after changing the sriov_numvfs value
        # for the changes to come into effect.
        # Any identified card/vendor can then be special-cased here, if
        # needed.
        pass


def apply_vlan_filter_for_vf(pf, vf, vlan_name, vlan_id, prefix='/'):
    """
    Apply the hardware VLAN filtering for the selected VF.
    """

    # this is more complicated, because to do this, we actually need to have
    # the vf index - just knowing the vf interface name is not enough
    vf_index = None
    # the prefix argument is here only for unit testing purposes
    vf_devdir = os.path.join(prefix, 'sys/class/net', vf, 'device')
    vf_dev_id = os.path.basename(os.readlink(vf_devdir))
    pf_devdir = os.path.join(prefix, 'sys/class/net', pf, 'device')
    for f in os.listdir(pf_devdir):
        if 'virtfn' in f:
            dev_path = os.path.join(pf_devdir, f)
            dev_id = os.path.basename(os.readlink(dev_path))
            if dev_id == vf_dev_id:
                vf_index = f[6:]
                break

    if not vf_index:
        raise RuntimeError(
            'could not determine the VF index for %s while configuring vlan %s' % (vf, vlan_name))

    # now, create the VLAN filter
    # TODO: would be best if we did this directl via python, without calling
    #  the iproute tooling
    try:
        subprocess.check_call(['ip', 'link', 'set',
                               'dev', pf,
                               'vf', vf_index,
                               'vlan', str(vlan_id)],
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        raise RuntimeError(
            'failed setting SR-IOV VLAN filter for vlan %s (ip link set command failed)' % vlan_name)


def apply_sriov_config(config_manager, rootdir='/'):
    """
    Go through all interfaces, identify which ones are SR-IOV VFs, create
    them and perform all other necessary setup.
    """
    config_manager.parse()
    interfaces = utils.get_interfaces()
    np_state = config_manager.np_state

    # for sr-iov devices, we identify VFs by them having a link: field
    # pointing to an PF. So let's browse through all ethernet devices,
    # find all that are VFs and count how many of those are linked to
    # particular PFs, as we need to then set the numvfs for each.
    vf_counts = _get_vf_number_per_pf(np_state)
    # we also store all matches between VF/PF netplan entry names and
    # interface that they're currently matching to
    vfs_set = _get_virtual_functions(np_state)
    pfs = _get_physical_functions(np_state)

    # setup the required number of VFs per PF
    # at the same time store which PFs got changed in case the NICs
    # require some special quirks for the VF number to change
    vf_count_changed = []
    if vf_counts:
        for pf, vf_count in vf_counts.items():
            if not set_numvfs_for_pf(pf, vf_count):
                continue

            vf_count_changed.append(pf)

    if vf_count_changed:
        # some cards need special treatment when we want to change the
        # number of enabled VFs
        for pf in vf_count_changed:
            perform_hardware_specific_quirks(pf)

        # also, since the VF number changed, the interfaces list also
        # changed, so we need to refresh it
        interfaces = utils.get_interfaces()

    # now in theory we should have all the new VFs set up and existing;
    # this is needed because we will have to now match the defined VF
    # entries to existing interfaces, otherwise we won't be able to set
    # filtered VLANs for those.
    # XXX: does matching those even make sense?
    vfs = {}
    for vf in vfs_set:
        netdef = np_state[vf]
        if netdef._has_match:
            # right now we only match by name, as I don't think matching per
            # driver and/or macaddress makes sense
            # TODO: print warning if other matches are provided

            for interface in interfaces:
                if netdef._match_interface(iface_name=interface):
                    if vf in vfs and vfs[vf]:
                        raise ConfigurationError('matched more than one interface for a VF device: %s' % vf)
                    vfs[vf] = interface
        else:
            if vf in interfaces:
                vfs[vf] = vf

    # Walk the SR-IOV PFs and check if we need to change the eswitch mode
    for netdef_id, iface in pfs.items():
        netdef = np_state[netdef_id]
        eswitch_mode = netdef._embedded_switch_mode
        if eswitch_mode in ['switchdev', 'legacy']:
            pci_addr = _get_pci_slot_name(iface)
            pcidev = PCIDevice(pci_addr)
            current_eswitch_mode_system = pcidev.devlink_eswitch_mode()
            if eswitch_mode != current_eswitch_mode_system:
                if pcidev.is_pf:
                    logging.debug("Found VFs of {}: {}".format(pcidev, pcidev.vf_addrs))
                    if pcidev.vfs:
                        try:
                            unbind_vfs(pcidev.vfs, pcidev.driver)
                        except Exception as e:
                            logging.warning(f'Unbinding of VFs for {netdef_id} failed: {str(e)}')

                    logging.debug(f'Changing eswitch mode from {current_eswitch_mode_system} to {eswitch_mode} for: {netdef_id}')
                    pcidev.devlink_set('eswitch', 'mode', eswitch_mode)

                    if pcidev.vfs:
                        if not netdef._delay_virtual_functions_rebind:
                            bind_vfs(pcidev.vfs, pcidev.driver)

    filtered_vlans_set = set()
    for vlan, netdef in np_state.vlans.items():
        # there is a special sriov vlan renderer that one can use to mark
        # a selected vlan to be done in hardware (VLAN filtering)
        if netdef._has_sriov_vlan_filter:
            # this only works for SR-IOV VF interfaces
            link = netdef.links.get('vlan')
            vlan_id = netdef._vlan_id

            vf = vfs.get(link.id)
            if not vf:
                # it is possible this is not an error, for instance when
                # the configuration has been defined 'for the future'
                # XXX: but maybe we should error out here as well?
                logging.warning(
                    'SR-IOV vlan defined for %s but link %s is either not a VF or has no matches' % (vlan, link.id))
                continue

            # get the parent pf interface
            # first we fetch the related vf netplan entry
            # and finally, get the matched pf interface
            pf = pfs.get(link.links.get('sriov').id)

            if vf in filtered_vlans_set:
                raise ConfigurationError(
                    'interface %s for netplan device %s (%s) already has an SR-IOV vlan defined' % (vf, link.id, vlan))

            # TODO: make sure that we don't apply the filter twice
            apply_vlan_filter_for_vf(pf, vf, vlan, vlan_id)
            filtered_vlans_set.add(vf)



#
# Functional tests of netplan's keyfile parser that verify that the generated
# YAML files look as expected. These are run during "make check" and
# don't touch the system configuration at all.
#
# Copyright (C) 2021 Canonical, Ltd.
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from configparser import ConfigParser
import netplan
import os
import re
import sys
import shutil
import tempfile
import unittest
import contextlib
import subprocess

exe_configure = os.environ.get('NETPLAN_CONFIGURE_PATH',
                               os.path.join(os.path.dirname(os.path.dirname(
                                            os.path.dirname(os.path.abspath(__file__)))), 'configure'))

# make sure we point to libnetplan properly.
os.environ.update({'LD_LIBRARY_PATH': '.:{}'.format(os.environ.get('LD_LIBRARY_PATH'))})

# make sure we fail on criticals
os.environ['G_DEBUG'] = 'fatal-criticals'

WOKE_REPLACE_REGEX = ' +# wokeignore:rule=[a-z]+'


# A contextmanager to catch the output on a low level so that it catches output
# from a subprocess or C library call, in addition to normal python output
@contextlib.contextmanager
def capture_stderr():
    stderr_fd = 2  # 2 = stderr
    with tempfile.NamedTemporaryFile(mode='w+b') as tmp:
        stderr_copy = os.dup(stderr_fd)
        try:
            sys.stderr.flush()
            os.dup2(tmp.fileno(), stderr_fd)
            yield tmp
        finally:
            sys.stderr.flush()
            os.dup2(stderr_copy, stderr_fd)
            os.close(stderr_copy)


class TestKeyfileBase(unittest.TestCase):

    def setUp(self):
        self.workdir = tempfile.TemporaryDirectory()
        self.confdir = os.path.join(self.workdir.name, 'etc', 'netplan')
        self.maxDiff = None
        os.makedirs(self.confdir)

    def tearDown(self):
        shutil.rmtree(self.workdir.name)
        super().tearDown()

    def generate_from_keyfile(self, keyfile, netdef_id=None, expect_fail=False, filename=None, regenerate=True):
        '''Call libnetplan with given keyfile string as configuration'''
        # Autodetect default 'NM-<UUID>' netdef-id
        ssid = ''
        keyfile = re.sub(WOKE_REPLACE_REGEX, '', keyfile)
        # calculate the UUID+SSID string
        found_values = 0
        uuid = 'UNKNOWN_UUID'
        ssid = ''
        for line in keyfile.splitlines():
            if line.startswith('uuid='):
                uuid = line.split('=')[1]
                found_values += 1
            elif line.startswith('ssid='):
                ssid += '-' + line.split('=')[1]
                found_values += 1
            if found_values >= 2:
                break
        if not netdef_id:
            netdef_id = 'NM-' + uuid
        yaml_path = os.path.join(self.workdir.name, 'etc', 'netplan', '90-NM-'+uuid+'.yaml')
        generated_file = 'netplan-{}{}.nmconnection'.format(netdef_id, ssid)
        original_file = filename or generated_file
        f = os.path.join(self.workdir.name,
                         'run/NetworkManager/system-connections/{}'.format(original_file))
        os.makedirs(os.path.dirname(f))
        # Create the original keyfile that will be parsed by netplan
        with open(f, 'w') as file:
            file.write(keyfile)

        with capture_stderr() as outf:
            parser = netplan.Parser()
            if expect_fail:
                try:
                    parser.load_keyfile(f)
                except netplan.NetplanException as err:
                    return err.message
            else:
                ret = parser.load_keyfile(f)  # Throws netplan.NetplanExcption on failure
                self.assertTrue(ret)
                # If the original file does not have a standard netplan-*.nmconnection
                # filename it is being deleted in favor of the newly generated file.
                # It has been parsed and is not needed anymore in this case
                if generated_file != original_file:
                    os.remove(f)
                state = netplan.State()
                state.import_parser_results(parser)
                with open(yaml_path, 'w') as f:
                    os.chmod(yaml_path, mode=0o600)
                    state._dump_yaml(f)
                # check re-generated keyfile
                if regenerate:
                    self.assert_nm_regenerate({generated_file: keyfile})
            with open(outf.name, 'r') as f:
                output = f.read().strip()  # output from stderr (fd=2) on C/library level
                return output

    def assert_netplan(self, file_contents_map):
        for uuid in file_contents_map.keys():
            file_contents_map[uuid] = re.sub(WOKE_REPLACE_REGEX, '', file_contents_map[uuid])
            path = os.path.join(self.confdir, '90-NM-{}.yaml'.format(uuid))
            self.assertTrue(os.path.isfile(path))
            st = os.stat(path)
            permission = oct(st.st_mode & 0o777)
            self.assertEqual(permission, '0o600')
            with open(os.path.join(self.confdir, '90-NM-{}.yaml'.format(uuid)), 'r') as f:
                self.assertEqual(f.read(), file_contents_map[uuid])

    def normalize_keyfile(self, file_contents):
        parser = ConfigParser()
        parser.read_string(file_contents)
        sections = parser.sections()
        res = []
        # Sort sections and keys
        sections.sort()
        for s in sections:
            items = parser.items(s)
            if s == 'ipv6' and len(items) == 1 and items[0] == ('method', 'ignore'):
                continue

            line = '\n[' + s + ']'
            res.append(line)
            items.sort(key=lambda tup: tup[0])
            for k, v in items:
                # Normalize lines
                if k == 'addr-gen-mode':
                    v = v.replace('1', 'stable-privacy').replace('0', 'eui64')
                elif k == 'ip6-privacy' and v == '0':
                    continue
                elif k == 'wake-on-lan' and v == '1':
                    continue
                elif k == 'stp' and v == 'true':
                    continue
                elif k.startswith('route'):
                    v = v.replace(',::', ',').replace(',0.0.0.0', ',')
                    v = v.strip(',')

                line = (k + '=' + v).strip(';')
                res.append(line)
        return '\n'.join(res).strip()+'\n'

    def assert_nm_regenerate(self, file_contents_map):
        argv = [exe_configure, '--root-dir', self.workdir.name, '--networkmanager-only']
        p = subprocess.Popen(argv, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, text=True)
        returncode = p.wait(5)
        (out, err) = p.communicate()
        self.assertEqual(returncode, 0, err)
        self.assertEqual(out, '')
        con_dir = os.path.join(self.workdir.name, 'run', 'NetworkManager', 'system-connections')
        if file_contents_map:
            self.assertEqual(set(os.listdir(con_dir)),
                             set([n for n in file_contents_map]))
            for fname, contents in file_contents_map.items():
                contents = re.sub(WOKE_REPLACE_REGEX, '', contents)
                with open(os.path.join(con_dir, fname)) as f:
                    generated_keyfile = self.normalize_keyfile(f.read())
                    normalized_contents = self.normalize_keyfile(contents)
                    self.assertEqual(generated_keyfile, normalized_contents,
                                     'Re-generated keyfile does not match')
        else:  # pragma: nocover (only needed for test debugging)
            if os.path.exists(con_dir):
                self.assertEqual(os.listdir(con_dir), [])
        return err



#!/usr/bin/python3
# Dummy devices integration tests.          wokeignore:rule=dummy
#
# These need to be run in a VM and do change the system
# configuration.
#
# Copyright (C) 2023 Canonical, Ltd.
# Author: Danilo Egea Gondolfo <danilo.egea.gondolfo@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import subprocess
import unittest

from base import IntegrationTestsBase, test_backends


class _CommonTests():

    def test_create_single_interface(self):
        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'dm0'], stderr=subprocess.DEVNULL)
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  version: 2
  dummy-devices:        # wokeignore:rule=dummy
    dm0: {}
''' % {'r': self.backend})
        self.generate_and_settle(['dm0'])
        self.assert_iface('dm0')

    def test_create_multiple_interfaces(self):
        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'dm0'], stderr=subprocess.DEVNULL)
        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'dm1'], stderr=subprocess.DEVNULL)
        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'dm2'], stderr=subprocess.DEVNULL)
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  version: 2
  dummy-devices:        # wokeignore:rule=dummy
    dm0: {}
    dm1: {}
    dm2: {}
''' % {'r': self.backend})
        self.generate_and_settle(['dm0', 'dm1', 'dm2'])
        self.assert_iface('dm0')
        self.assert_iface('dm1')
        self.assert_iface('dm2')

    def test_interface_with_ip_addresses(self):
        self.addCleanup(subprocess.call, ['ip', 'link', 'delete', 'dm0'], stderr=subprocess.DEVNULL)
        with open(self.config, 'w') as f:
            f.write('''network:
  renderer: %(r)s
  version: 2
  dummy-devices:            # wokeignore:rule=dummy
    dm0:
      addresses:
        - 192.168.123.123/24
        - 1234:FFFF::42/64
''' % {'r': self.backend})
        self.generate_and_settle(['dm0'])
        self.assert_iface('dm0', ['inet 192.168.123.123/24', 'inet6 1234:ffff::42/64'])


@unittest.skipIf("networkd" not in test_backends,
                 "skipping as networkd backend tests are disabled")
class TestNetworkd(IntegrationTestsBase, _CommonTests):
    backend = 'networkd'


@unittest.skipIf("NetworkManager" not in test_backends,
                 "skipping as NetworkManager backend tests are disabled")
class TestNetworkManager(IntegrationTestsBase, _CommonTests):
    backend = 'NetworkManager'


unittest.main(testRunner=unittest.TextTestRunner(stream=sys.stdout, verbosity=2))



#!/usr/bin/python3
# Closed-box tests of netplan CLI. These are run during "make check" and don't
# touch the system configuration at all.
#
# Copyright (C) 2023 Canonical, Ltd.
# Authors: Lukas Märdian <slyon@ubuntu.com>
#          Danilo Egea Gondolfo <danilo.egea.gondolfo@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import json
import os
import shutil
import subprocess
import tempfile
import unittest
from unittest.mock import call, mock_open, patch

import yaml

from netplan_cli.cli.state import (Interface, NetplanConfigState,
                                   SystemConfigState)

from .test_status import (BRIDGE, DNS_ADDRESSES, DNS_IP4, DNS_SEARCH, FAKE_DEV,
                          IPROUTE2, NETWORKD, NMCLI, ROUTE4, ROUTE6)


class TestSystemState(unittest.TestCase):
    '''Test netplan state module'''

    def setUp(self):
        self.maxDiff = None

    @patch('subprocess.check_output')
    def test_query_iproute2(self, mock):
        mock.return_value = IPROUTE2
        res = SystemConfigState.query_iproute2()
        mock.assert_called_with(['ip', '-d', '-j', 'addr'], text=True)
        self.assertEqual(len(res), 7)
        self.assertListEqual([itf.get('ifname') for itf in res],
                             ['lo', 'enp0s31f6', 'wlan0', 'wg0', 'wwan0', 'tun0', 'tun1'])

    @patch('subprocess.check_output')
    def test_query_iproute2_fail(self, mock):
        mock.side_effect = subprocess.CalledProcessError(1, '', 'ERR')
        with self.assertLogs() as cm:
            res = SystemConfigState.query_iproute2()
            mock.assert_called_with(['ip', '-d', '-j', 'addr'], text=True)
            self.assertIsNone(res)
            self.assertIn('CRITICAL:root:Cannot query iproute2 interface data:', cm.output[0])

    @patch('subprocess.check_output')
    def test_query_networkd(self, mock):
        mock.return_value = NETWORKD
        res = SystemConfigState.query_networkd()
        mock.assert_called_with(['networkctl', '--json=short'], text=True)
        self.assertEqual(len(res), 10)
        self.assertListEqual([itf.get('Name') for itf in res],
                             ['lo', 'enp0s31f6', 'wlan0', 'wg0', 'wwan0', 'tun0', 'mybr0', 'mybond0', 'myvrf0', 'tun1'])

    @patch('subprocess.check_output')
    def test_query_networkd_fail(self, mock):
        mock.side_effect = subprocess.CalledProcessError(1, '', 'ERR')
        with self.assertLogs() as cm:
            res = SystemConfigState.query_networkd()
            mock.assert_called_with(['networkctl', '--json=short'], text=True)
            self.assertIsNone(res)
            self.assertIn('CRITICAL:root:Cannot query networkd interface data:', cm.output[0])

    @patch('subprocess.check_output')
    def test_query_nm(self, mock):
        mock.return_value = NMCLI
        res = SystemConfigState.query_nm()
        mock.assert_called_with(['nmcli', '-t', '-f',
                                 'DEVICE,NAME,UUID,FILENAME,TYPE,AUTOCONNECT',
                                 'con', 'show'], text=True)
        self.assertEqual(len(res), 2)
        self.assertListEqual([itf.get('device') for itf in res], ['wlan0', 'wlan1'])

    @patch('subprocess.check_output')
    def test_query_nm_fail(self, mock):
        mock.side_effect = subprocess.CalledProcessError(1, '', 'ERR')
        with self.assertLogs(level='DEBUG') as cm:
            res = SystemConfigState.query_nm()
            mock.assert_called_with(['nmcli', '-t', '-f',
                                     'DEVICE,NAME,UUID,FILENAME,TYPE,AUTOCONNECT',
                                     'con', 'show'], text=True)
            self.assertIsNone(res)
            self.assertIn('DEBUG:root:Cannot query NetworkManager interface data:', cm.output[0])

    @patch('subprocess.check_output')
    def test_query_routes(self, mock):
        mock.side_effect = [ROUTE4, ROUTE6]
        res4, res6 = SystemConfigState.query_routes()
        mock.assert_has_calls([
            call(['ip', '-d', '-j', '-4', 'route', 'show', 'table', 'all'], text=True),
            call(['ip', '-d', '-j', '-6', 'route', 'show', 'table', 'all'], text=True),
            ])
        self.assertEqual(len(res4), 7)
        self.assertListEqual([route.get('dev') for route in res4],
                             ['enp0s31f6', 'wlan0', 'wg0', 'enp0s31f6', 'wlan0', 'enp0s31f6', 'enp0s31f6'])
        self.assertEqual(len(res6), 10)
        self.assertListEqual([route.get('dev') for route in res6],
                             ['lo', 'enp0s31f6', 'wlan0', 'enp0s31f6', 'wlan0',
                              'tun0', 'enp0s31f6', 'wlan0', 'enp0s31f6', 'wlan0'])

    @patch('subprocess.check_output')
    def test_query_routes_fail(self, mock):
        mock.side_effect = subprocess.CalledProcessError(1, '', 'ERR')
        with self.assertLogs(level='DEBUG') as cm:
            res4, res6 = SystemConfigState.query_routes()
            mock.assert_called_with(['ip', '-d', '-j', '-4', 'route', 'show', 'table', 'all'], text=True)
            self.assertIsNone(res4)
            self.assertIsNone(res6)
            self.assertIn('DEBUG:root:Cannot query iproute2 route data:', cm.output[0])

    @patch('subprocess.check_output')
    def test_query_resolved(self, mock_busctl):
        mock_busctl.return_value = '''{"data":[{
        "DNS": {
            "type": "a(iiay)",
            "data": '''+json.dumps(DNS_ADDRESSES)+'''
        },
        "Domains": {
            "type": "a(isb)",
            "data": '''+json.dumps(DNS_SEARCH)+'''
        }}]}'''
        addresses, search = SystemConfigState.query_resolved()
        self.assertEqual(len(addresses), 4)
        self.assertListEqual([addr[0] for addr in addresses],
                             [5, 5, 2, 2])  # interface index
        self.assertEqual(len(search), 2)
        self.assertListEqual([s[1] for s in search],
                             ['search.domain', 'search.domain'])

    @patch('subprocess.check_output')
    def test_query_resolved_fail(self, mock_busctl):
        mock_busctl.return_value = '{"data":[{"DNS":{"type":"invalid","data":"garbage"}}]}'
        with self.assertLogs(level='DEBUG') as cm:
            addresses, search = SystemConfigState.query_resolved()
            self.assertIsNone(addresses)
            self.assertIsNone(search)
            self.assertIn('DEBUG:root:Cannot query resolved DNS data: DNS address type doesn\'t match', cm.output[0])

    @patch('shutil.which')
    def test_query_resolved_fail_missing_busctl(self, mock):
        mock.return_value = None
        with self.assertLogs(level='DEBUG') as cm:
            addresses, search = SystemConfigState.query_resolved()
            self.assertIsNone(addresses)
            self.assertIsNone(search)
            self.assertIn('DEBUG:root:Cannot query resolved DNS data: missing busctl utility', cm.output[0])

    def test_query_resolvconf(self):
        with patch('builtins.open', mock_open(read_data='''\
nameserver 1.1.1.1
nameserver 8.8.8.8
options edns0 trust-ad
search   some.domain
search search.domain  another.one
''')):
            res = SystemConfigState.resolvconf_json()
            print(res)
            self.assertListEqual(res.get('addresses'), ['1.1.1.1', '8.8.8.8'])
            self.assertListEqual(res.get('search'), ['search.domain', 'another.one'])
            self.assertEqual(res.get('mode'), None)

    def test_query_resolvconf_stub(self):
        with patch('builtins.open', mock_open(read_data='\
# This is /run/systemd/resolve/stub-resolv.conf managed by man:systemd-resolved(8).')):
            res = SystemConfigState.resolvconf_json()
            self.assertEqual(res.get('mode'), 'stub')

    def test_query_resolvconf_compat(self):
        with patch('builtins.open', mock_open(read_data='\
# This is /run/systemd/resolve/resolv.conf managed by man:systemd-resolved(8).')):
            res = SystemConfigState.resolvconf_json()
            self.assertEqual(res.get('mode'), 'compat')

    def test_query_resolvconf_fail(self):
        with self.assertLogs() as cm:
            with patch('builtins.open', mock_open(read_data='')) as mock_file:
                mock_file.side_effect = Exception(1, '', 'ERR')
                SystemConfigState.resolvconf_json()
                self.assertIn('WARNING:root:Cannot parse /etc/resolv.conf:', cm.output[0])

    def test_query_online_state_online(self):
        dev = copy.deepcopy(FAKE_DEV)
        dev['addr_info'] = [{
            'local': '192.168.0.100',
            'prefixlen': 24,
        }]
        dev['flags'].append('UP')
        dev['operstate'] = 'UP'
        routes = [{
            'dst': 'default',
            'gateway': '192.168.0.1',
            'dev': dev['ifname'],
        }]
        dns = [(FAKE_DEV['ifindex'], 2, DNS_IP4)]
        res = SystemConfigState.query_online_state([Interface(dev, [], [], (dns, None), (routes, None))])
        self.assertTrue(res)

    def test_query_online_state_offline(self):
        res = SystemConfigState.query_online_state([Interface(FAKE_DEV, [])])
        self.assertFalse(res)

    @patch('netplan_cli.cli.utils.systemctl')
    @patch('netplan_cli.cli.state.SystemConfigState.query_members')
    @patch('netplan_cli.cli.state.SystemConfigState.query_iproute2')
    @patch('netplan_cli.cli.state.SystemConfigState.query_networkd')
    @patch('netplan_cli.cli.state.SystemConfigState.query_nm')
    @patch('netplan_cli.cli.state.SystemConfigState.query_routes')
    @patch('netplan_cli.cli.state.SystemConfigState.query_resolved')
    @patch('netplan_cli.cli.state.SystemConfigState.resolvconf_json')
    @patch('netplan_cli.cli.state.SystemConfigState.query_online_state')
    def test_system_state_config_data_interfaces(self, online_mock, resolvconf_mock, rd_mock,
                                                 routes_mock, nm_mock, networkd_mock, iproute2_mock,
                                                 members_mock, systemctl_mock):
        systemctl_mock.return_value = None
        members_mock.return_value = []
        iproute2_mock.return_value = [FAKE_DEV, BRIDGE]
        nm_mock.return_value = []
        routes_mock.return_value = (None, None)
        rd_mock.return_value = (None, None)
        resolvconf_mock.return_value = {'addresses': [], 'search': [], 'mode': None}
        online_mock.return_value = False
        networkd_mock.return_value = SystemConfigState.process_networkd(NETWORKD)
        state = SystemConfigState()
        self.assertIn('fakedev0', [iface.name for iface in state.interface_list])

    @patch('subprocess.check_output')
    def test_query_members(self, mock):
        mock.return_value = '[{"ifname":"eth0"}, {"ifname":"eth1"}]'
        members = SystemConfigState.query_members('mybr0')
        mock.assert_has_calls([
            call(['ip', '-d', '-j', 'link', 'show', 'master', 'mybr0'], text=True),     # wokeignore:rule=master
            ])
        self.assertListEqual(members, ['eth0', 'eth1'])

    @patch('subprocess.check_output')
    def test_query_members_fail(self, mock):
        mock.side_effect = subprocess.CalledProcessError(1, '', 'ERR')
        with self.assertLogs(level='WARNING') as cm:
            bridge = SystemConfigState.query_members('mybr0')
            mock.assert_has_calls([
                call(['ip', '-d', '-j', 'link', 'show', 'master', 'mybr0'], text=True),     # wokeignore:rule=master
                ])
            self.assertListEqual(bridge, [])
            self.assertIn('WARNING:root:Cannot query bridge:', cm.output[0])

    @patch('netplan_cli.cli.state.SystemConfigState.query_members')
    def test_correlate_members_and_uplink_bridge(self, mock):
        mock.side_effect = lambda _: ['eth0', 'eth1']
        interface1 = Interface({'ifname': 'eth0'})
        interface1.nd = {'Type': 'ether'}
        interface2 = Interface({'ifname': 'eth1'})
        interface2.nd = {'Type': 'ether'}
        interface3 = Interface({'ifname': 'br0'})
        interface3.nd = {'Type': 'bridge'}
        SystemConfigState.correlate_members_and_uplink([interface1, interface2, interface3])
        self.assertEqual(interface1.bridge, 'br0')
        self.assertEqual(interface2.bridge, 'br0')
        self.assertListEqual(interface3.members, ['eth0', 'eth1'])

    @patch('netplan_cli.cli.state.SystemConfigState.query_members')
    def test_correlate_members_and_uplink_bond(self, mock):
        mock.side_effect = lambda _: ['eth2', 'eth3']
        interface1 = Interface({'ifname': 'eth2'})
        interface1.nd = {'Type': 'ether'}
        interface2 = Interface({'ifname': 'eth3'})
        interface2.nd = {'Type': 'ether'}
        interface3 = Interface({'ifname': 'bond0'})
        interface3.nd = {'Type': 'bond'}
        SystemConfigState.correlate_members_and_uplink([interface1, interface2, interface3])
        self.assertEqual(interface1.bond, 'bond0')
        self.assertEqual(interface2.bond, 'bond0')
        self.assertListEqual(interface3.members, ['eth2', 'eth3'])

    @patch('netplan_cli.cli.state.SystemConfigState.query_members')
    def test_correlate_members_and_uplink_vrf(self, mock):
        mock.side_effect = lambda _: ['eth2', 'eth3']
        interface1 = Interface({'ifname': 'eth2'})
        interface1.nd = {'Type': 'ether'}
        interface2 = Interface({'ifname': 'eth3'})
        interface2.nd = {'Type': 'ether'}
        interface3 = Interface({'ifname': 'vrf0'})
        interface3.nd = {'Type': 'ether', 'Kind': 'vrf'}
        SystemConfigState.correlate_members_and_uplink([interface1, interface2, interface3])
        self.assertEqual(interface1.vrf, 'vrf0')
        self.assertEqual(interface2.vrf, 'vrf0')
        self.assertListEqual(interface3.members, ['eth2', 'eth3'])


class TestNetplanState(unittest.TestCase):
    '''Test netplan state NetplanConfigState class'''

    def setUp(self):
        self.workdir = tempfile.TemporaryDirectory(prefix='netplan_')
        self.file = '70-netplan-set.yaml'
        self.path = os.path.join(self.workdir.name, 'etc', 'netplan', self.file)
        os.makedirs(os.path.join(self.workdir.name, 'etc', 'netplan'))
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: true
  bridges:
    br0:
      dhcp4: true''')

    def tearDown(self):
        shutil.rmtree(self.workdir.name)

    def test_get_data(self):
        state = NetplanConfigState(rootdir=self.workdir.name)
        state_data = state.get_data()
        self.assertIn('eth0', state_data.get('network').get('ethernets'))
        self.assertIn('br0', state_data.get('network').get('bridges'))

    def test_get_data_subtree(self):
        state = NetplanConfigState(subtree='ethernets', rootdir=self.workdir.name)
        state_data = state.get_data()
        self.assertIn('eth0', state_data)
        self.assertNotIn('br0', state_data)


class TestInterface(unittest.TestCase):
    '''Test netplan state Interface class'''

    @patch('subprocess.check_output')
    def test_query_nm_ssid(self, mock):
        mock.return_value = ' MYSSID '  # added some whitespace to strip()
        con = 'SOME_CONNECTION_ID'
        itf = Interface(FAKE_DEV, [])
        res = itf.query_nm_ssid(con)
        mock.assert_called_with(['nmcli', '--get-values', '802-11-wireless.ssid',
                                 'con', 'show', 'id', con],
                                text=True)
        self.assertEqual(res, 'MYSSID')

    @patch('subprocess.check_output')
    def test_query_nm_ssid_fail(self, mock):
        mock.side_effect = subprocess.CalledProcessError(1, '', 'ERR')
        con = 'SOME_CONNECTION_ID'
        itf = Interface(FAKE_DEV, [])
        with self.assertLogs() as cm:
            res = itf.query_nm_ssid(con)
            mock.assert_called_with(['nmcli', '--get-values', '802-11-wireless.ssid',
                                     'con', 'show', 'id', con],
                                    text=True)
            self.assertIsNone(res)
            self.assertIn('WARNING:root:Cannot query NetworkManager SSID for {}:'.format(con), cm.output[0])

    @patch('subprocess.check_output')
    def test_query_networkctl(self, mock):
        mock.return_value = 'DOES NOT MATTER'
        dev = 'fakedev0'
        itf = Interface(FAKE_DEV, [])
        res = itf.query_networkctl(dev)
        mock.assert_called_with(['networkctl', 'status', '--', dev], text=True)
        self.assertEqual(res, mock.return_value)

    @patch('subprocess.check_output')
    def test_query_networkctl_fail(self, mock):
        mock.side_effect = subprocess.CalledProcessError(1, '', 'ERR')
        dev = 'fakedev0'
        itf = Interface(FAKE_DEV, [])
        with self.assertLogs() as cm:
            res = itf.query_networkctl(dev)
            mock.assert_called_with(['networkctl', 'status', '--', dev], text=True)
            self.assertIsNone(res)
            self.assertIn('WARNING:root:Cannot query networkctl for {}:'.format(dev), cm.output[0])

    @patch('netplan_cli.cli.state.Interface.query_nm_ssid')
    @patch('netplan_cli.cli.state.Interface.query_networkctl')
    def test_json_nm_wlan0(self, networkctl_mock, nm_ssid_mock):
        SSID = 'MYCON'
        nm_ssid_mock.return_value = SSID
        # networkctl mock output reduced to relevant lines
        networkctl_mock.return_value = \
            'WiFi access point: {} (b4:fb:e4:75:c6:21)'.format(SSID)

        data = next((itf for itf in yaml.safe_load(IPROUTE2) if itf['ifindex'] == 5), {})
        nd = SystemConfigState.process_networkd(NETWORKD)
        nm = SystemConfigState.process_nm(NMCLI)
        dns = (DNS_ADDRESSES, DNS_SEARCH)
        routes = (SystemConfigState.process_generic(ROUTE4), SystemConfigState.process_generic(ROUTE6))

        itf = Interface(data, nd, nm, dns, routes)
        self.assertTrue(itf.up)
        self.assertFalse(itf.down)
        ifname, json = itf.json()
        self.assertEqual(ifname, 'wlan0')
        self.assertEqual(json.get('index'), 5)
        self.assertEqual(json.get('macaddress'), '1c:4d:70:e4:e4:0e')
        self.assertEqual(json.get('type'), 'wifi')
        self.assertEqual(json.get('ssid'), 'MYCON')
        self.assertEqual(json.get('backend'), 'NetworkManager')
        self.assertEqual(json.get('id'), 'NM-b6b7a21d-186e-45e1-b3a6-636da1735563')
        self.assertEqual(json.get('vendor'), 'Intel Corporation')
        self.assertEqual(json.get('adminstate'), 'UP')
        self.assertEqual(json.get('operstate'), 'UP')
        self.assertEqual(len(json.get('addresses')), 4)
        self.assertEqual(len(json.get('dns_addresses')), 2)
        self.assertEqual(len(json.get('dns_search')), 1)
        self.assertEqual(len(json.get('routes')), 6)

    @patch('netplan_cli.cli.state.Interface.query_nm_ssid')
    @patch('netplan_cli.cli.state.Interface.query_networkctl')
    def test_json_nm_wlan0_2(self, networkctl_mock, nm_ssid_mock):
        SSID = 'MYCON'
        nm_ssid_mock.return_value = SSID
        # networkctl mock output reduced to relevant lines
        # Newer versions of systemd changed WiFi to Wi-Fi
        # See systemd commit 8fff78a1dded105e1ee87bc66e29ef2fd61bf8c9
        networkctl_mock.return_value = \
            'Wi-Fi access point: {} (b4:fb:e4:75:c6:21)'.format(SSID)

        data = next((itf for itf in yaml.safe_load(IPROUTE2) if itf['ifindex'] == 5), {})
        nd = SystemConfigState.process_networkd(NETWORKD)
        nm = SystemConfigState.process_nm(NMCLI)
        dns = (DNS_ADDRESSES, DNS_SEARCH)
        routes = (SystemConfigState.process_generic(ROUTE4), SystemConfigState.process_generic(ROUTE6))

        itf = Interface(data, nd, nm, dns, routes)
        _, json = itf.json()
        self.assertEqual(json.get('ssid'), 'MYCON')

    @patch('netplan_cli.cli.state.Interface.query_nm_ssid')
    @patch('netplan_cli.cli.state.Interface.query_networkctl')
    def test_json_nd_wlan1_non_ascii(self, networkctl_mock, nm_ssid_mock):
        ND_SSID = '\\303\\241\\303\\251\\303\\255\\303\\263\\303\\272\\302\\242\\302\\242\\302\\242\\302\\243\\302\\243\\302\\243'
        NM_SSID = 'áéíóú¢¢¢£££'
        nm_ssid_mock.return_value = NM_SSID
        # networkctl mock output reduced to relevant lines
        # Newer versions of systemd changed WiFi to Wi-Fi
        # See systemd commit 8fff78a1dded105e1ee87bc66e29ef2fd61bf8c9
        networkctl_mock.return_value = \
            'Wi-Fi access point: {} (b4:fb:e4:75:c6:21)'.format(ND_SSID)

        data = {'ifname': 'wlan1', 'ifindex': 123}
        nd = [{'Index': 123, 'Type': 'wlan', 'Name': 'wlan1', 'SetupState': 'managed',
               'NetworkFile': '/run/systemd/network/10-netplan-wlan1.network'}]
        nm = SystemConfigState.process_nm(NMCLI)

        itf = Interface(data, nd, nm, (None, None), (None, None))
        _, json = itf.json()
        self.assertEqual(json.get('ssid'), NM_SSID)

    @patch('netplan_cli.cli.state.Interface.query_nm_ssid')
    def test_json_nm_wlan1_non_ascii(self, nm_ssid_mock):
        NM_SSID = 'áéíóú¢¢¢£££'
        nm_ssid_mock.return_value = NM_SSID

        data = {'ifname': 'wlan1', 'ifindex': 123}
        nd = [{'Index': 123, 'Type': 'wlan', 'Name': 'wlan1', 'SetupState': 'unmanaged'}]
        nm = SystemConfigState.process_nm(NMCLI)

        itf = Interface(data, nd, nm, (None, None), (None, None))
        _, json = itf.json()
        self.assertEqual(json.get('ssid'), NM_SSID)

    @patch('netplan_cli.cli.state.Interface.query_networkctl')
    def test_json_nd_enp0s31f6(self, networkctl_mock):
        # networkctl mock output reduced to relevant lines
        networkctl_mock.return_value = 'Activation Policy: manual'

        data = next((itf for itf in yaml.safe_load(IPROUTE2) if itf['ifindex'] == 2), {})
        nd = SystemConfigState.process_networkd(NETWORKD)
        nm = SystemConfigState.process_nm(NMCLI)
        dns = (DNS_ADDRESSES, DNS_SEARCH)
        routes = (SystemConfigState.process_generic(ROUTE4), SystemConfigState.process_generic(ROUTE6))

        itf = Interface(data, nd, nm, dns, routes)
        self.assertTrue(itf.up)
        self.assertFalse(itf.down)
        ifname, json = itf.json()
        self.assertEqual(ifname, 'enp0s31f6')
        self.assertEqual(json.get('index'), 2)
        self.assertEqual(json.get('macaddress'), '54:e1:ad:5f:24:b4')
        self.assertEqual(json.get('type'), 'ethernet')
        self.assertEqual(json.get('backend'), 'networkd')
        self.assertEqual(json.get('id'), 'enp0s31f6')
        self.assertEqual(json.get('vendor'), 'Intel Corporation')
        self.assertEqual(json.get('adminstate'), 'UP')
        self.assertEqual(json.get('operstate'), 'UP')
        self.assertEqual(json.get('activation_mode'), 'manual')
        self.assertEqual(len(json.get('addresses')), 3)
        _, meta = list(json.get('addresses')[0].items())[0]  # get first (any only) address
        self.assertIn('dhcp', meta.get('flags'))
        self.assertEqual(len(json.get('dns_addresses')), 2)
        self.assertEqual(len(json.get('dns_search')), 1)
        self.assertEqual(len(json.get('routes')), 8)

    def test_json_nd_tunnel(self):
        data = next((itf for itf in yaml.safe_load(IPROUTE2) if itf['ifindex'] == 41), {})
        nd = SystemConfigState.process_networkd(NETWORKD)

        itf = Interface(data, nd, [], (None, None), (None, None))
        ifname, json = itf.json()
        self.assertEqual(ifname, 'wg0')
        self.assertEqual(json.get('index'), 41)
        self.assertEqual(json.get('type'), 'tunnel')
        self.assertEqual(json.get('backend'), 'networkd')
        self.assertEqual(json.get('tunnel_mode'), 'wireguard')

    def test_json_no_type_id_backend(self):
        itf = Interface(FAKE_DEV, [], [], (None, None), (None, None))
        ifname, json = itf.json()
        self.assertEqual(ifname, 'fakedev0')
        self.assertEqual(json.get('index'), 42)
        self.assertNotIn('type', json)
        self.assertNotIn('id', json)
        self.assertNotIn('backend', json)

    def test_find_data_sources(self):
        nd_data = [{
            'Index': 42,
            'DNS': [
                {
                    'Address': [192, 168, 0, 1],
                    'ConfigSource': 'DHCPv4',
                },
                {
                    'Address': [0xab, 0xcd, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x12, 0x34],
                    'ConfigSource': 'DHCPv6',
                },
            ],
            'SearchDomains': [
                {
                    'Domain': 'mydomain.local',
                    'ConfigSource': 'DHCPv6',
                },
            ],
            'Addresses': [
                {
                    'Address': [192, 168, 0, 254],
                    'PrefixLength': 24,
                    'ConfigSource': 'DHCPv4',
                },
                {
                    'Address': [0xab, 0xcd, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x56, 0x78],
                    'PrefixLength': 64,
                    'ConfigSource': 'DHCPv6',
                },
            ],
        }]

        fake_dev = {
            'ifindex': 42,
            'ifname': 'fakedev0',
            'flags': [],
            'operstate': 'DOWN',
            'addr_info': [
                {
                    'local': '192.168.0.254',
                    'prefixlen': 24,
                    'dynamic': True
                },
                {
                    'local': 'abcd::5678',
                    'prefixlen': 64,
                    'dynamic': True
                }
            ]
        }
        itf = Interface(fake_dev, nd_data, [], (None, None), (None, None))

        expected = {
            'dns': {
                '192.168.0.1': 'DHCPv4',
                'abcd::1234': 'DHCPv6'
            },
            'search': {
                'mydomain.local': 'DHCPv6'
            },
            'addresses': {
                '192.168.0.254/24': 'DHCPv4',
                'abcd::5678/64': 'DHCPv6'}
        }
        self.assertDictEqual(expected, itf.data_sources)



#!/usr/bin/python3
# Functional tests of NetworkManager keyfile parser. These are run during
# "make check" and don't touch the system configuration at all.
#
# Copyright (C) 2021 Canonical, Ltd.
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import ctypes
import ctypes.util

import netplan

from .base import TestKeyfileBase

rootdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
exe_cli = os.path.join(rootdir, 'src', 'netplan.script')

lib = ctypes.CDLL('libnetplan.so.1')
UUID = 'ff9d6ebc-226d-4f82-a485-b7ff83b9607f'


class TestNetworkManagerKeyfileParser(TestKeyfileBase):
    '''Test NM keyfile parser as used by NetworkManager's YAML backend'''

    def test_keyfile_missing_uuid(self):
        err = self.generate_from_keyfile('[connection]\ntype=ethernets', expect_fail=True)
        self.assertIn('netplan: Keyfile: cannot find connection.uuid', err)

    def test_keyfile_missing_type(self):
        err = self.generate_from_keyfile('[connection]\nuuid=87749f1d-334f-40b2-98d4-55db58965f5f', expect_fail=True)
        self.assertIn('netplan: Keyfile: cannot find connection.type', err)

    def test_keyfile_gsm(self):
        self.generate_from_keyfile('''[connection]
id=T-Mobile Funkadelic 2
uuid={}
type=gsm

[gsm]
apn=internet2.voicestream.com
device-id=da812de91eec16620b06cd0ca5cbc7ea25245222
home-only=true
network-id=254098
password=parliament2
pin=123456
sim-id=89148000000060671234
sim-operator-id=310260
username=george.clinton.again
mtu=1042

[ipv4]
dns-search=
method=auto

[ipv6]
dns-search=
method=auto
ip6-privacy=0
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  modems:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      dhcp4: true
      dhcp6: true
      mtu: 1042
      apn: "internet2.voicestream.com"
      device-id: "da812de91eec16620b06cd0ca5cbc7ea25245222"
      network-id: "254098"
      pin: "123456"
      sim-id: "89148000000060671234"
      sim-operator-id: "310260"
      username: "george.clinton.again"
      password: "parliament2"
      networkmanager:
        uuid: "{}"
        name: "T-Mobile Funkadelic 2"
        passthrough:
          gsm.home-only: "true"
          ipv4.dns-search: ""
          ipv6.dns-search: ""
'''.format(UUID, UUID)})

    def test_keyfile_cdma(self):
        self.generate_from_keyfile('''[connection]
id=T-Mobile Funkadelic 2
uuid={}
type=cdma

[cdma]
number=0123456
username=testuser
password=testpass
mtu=1042

[ipv4]
method=auto

[ipv6]
method=ignore
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  modems:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      dhcp4: true
      mtu: 1042
      username: "testuser"
      password: "testpass"
      number: "0123456"
      networkmanager:
        uuid: "{}"
        name: "T-Mobile Funkadelic 2"
'''.format(UUID, UUID)})

    def test_keyfile_gsm_via_bluetooth(self):
        self.generate_from_keyfile('''[connection]
id=T-Mobile Funkadelic 2
uuid={}
type=bluetooth

[gsm]
apn=internet2.voicestream.com
device-id=da812de91eec16620b06cd0ca5cbc7ea25245222
home-only=true
network-id=254098
password=parliament2
pin=123456
sim-id=89148000000060671234
sim-operator-id=310260
username=george.clinton.again

[ipv4]
dns-search=
method=auto

[ipv6]
dns-search=
method=auto

[proxy]'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  nm-devices:
    NM-{}:
      renderer: NetworkManager
      networkmanager:
        uuid: "{}"
        name: "T-Mobile Funkadelic 2"
        passthrough:
          connection.type: "bluetooth"
          gsm.apn: "internet2.voicestream.com"
          gsm.device-id: "da812de91eec16620b06cd0ca5cbc7ea25245222"
          gsm.home-only: "true"
          gsm.network-id: "254098"
          gsm.password: "parliament2"
          gsm.pin: "123456"
          gsm.sim-id: "89148000000060671234"
          gsm.sim-operator-id: "310260"
          gsm.username: "george.clinton.again"
          ipv4.dns-search: ""
          ipv4.method: "auto"
          ipv6.dns-search: ""
          ipv6.method: "auto"
          proxy._: ""
'''.format(UUID, UUID)})

    def test_keyfile_method_auto(self):
        self.generate_from_keyfile('''[connection]
id=Test
uuid={}
type=ethernet

[ethernet]
wake-on-lan=0
mtu=1500
cloned-mac-address=00:11:22:33:44:55

[ipv4]
dns-search=
method=auto
ignore-auto-routes=true
never-default=true
route-metric=4242

[ipv6]
addr-gen-mode=eui64
dns-search=
method=auto
ip6-privacy=0
ignore-auto-routes=true
never-default=true
route-metric=4242

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      dhcp4: true
      dhcp4-overrides:
        use-routes: false
        route-metric: 4242
      dhcp6: true
      dhcp6-overrides:
        use-routes: false
        route-metric: 4242
      macaddress: "00:11:22:33:44:55"
      ipv6-address-generation: "eui64"
      mtu: 1500
      networkmanager:
        uuid: "{}"
        name: "Test"
        passthrough:
          ipv4.dns-search: ""
          ipv6.dns-search: ""
          proxy._: ""
'''.format(UUID, UUID)})

    def test_keyfile_fail_validation(self):
        err = self.generate_from_keyfile('''[connection]
id=Test
uuid={}
type=ethernet

[ethernet]
wake-on-lan=0

[ipv4]
method=auto

[ipv6]
addr-gen-mode=eui64
token=::42
method=auto
'''.format(UUID), expect_fail=True)
        self.assertIn('Error in network definition:', err)

    def test_keyfile_method_manual(self):
        self.generate_from_keyfile('''[connection]
id=Test
uuid={}
type=ethernet

[ethernet]
mac-address=00:11:22:33:44:55

[ipv4]
dns-search=foo.local;bar.remote;
dns=9.8.7.6;5.4.3.2
method=manual
address1=1.2.3.4/24,8.8.8.8
address2=5.6.7.8/16
gateway=6.6.6.6
route1=1.1.2.2/16,8.8.8.8,42
route1_options=onlink=true,initrwnd=33,initcwnd=44,mtu=1024,table=102,src=10.10.10.11
route2=2.2.3.3/24,4.4.4.4

[ipv6]
addr-gen-mode=stable-privacy
dns-search=bar.local
dns=dead:beef::2;
method=manual
ip6-privacy=2
address1=1:2:3::9/128
gateway=6:6::6
route1=dead:beef::1/128,2001:1234::2
route1_options=unknown=invalid,
route2=4:5:6:7:8:9:0:1/63,,5

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        macaddress: "00:11:22:33:44:55"
      addresses:
      - "1.2.3.4/24"
      - "5.6.7.8/16"
      - "1:2:3::9/128"
      nameservers:
        addresses:
        - 9.8.7.6
        - 5.4.3.2
        - dead:beef::2
      gateway4: 6.6.6.6
      gateway6: 6:6::6
      ipv6-address-generation: "stable-privacy"
      ipv6-privacy: true
      routes:
      - metric: 42
        table: 102
        mtu: 1024
        congestion-window: 44
        advertised-receive-window: 33
        on-link: true
        from: "10.10.10.11"
        to: "1.1.2.2/16"
        via: "8.8.8.8"
      - to: "2.2.3.3/24"
        via: "4.4.4.4"
      - to: "dead:beef::1/128"
        via: "2001:1234::2"
      - scope: "link"
        metric: 5
        to: "4:5:6:7:8:9:0:1/63"
      wakeonlan: true
      networkmanager:
        uuid: "{}"
        name: "Test"
        passthrough:
          ipv4.dns-search: "foo.local;bar.remote;"
          ipv4.method: "manual"
          ipv4.address1: "1.2.3.4/24,8.8.8.8"
          ipv6.dns-search: "bar.local"
          ipv6.route1: "dead:beef::1/128,2001:1234::2"
          ipv6.route1_options: "unknown=invalid,"
          proxy._: ""
'''.format(UUID, UUID)})

    def test_keyfile_dummy(self):       # wokeignore:rule=dummy
        self.generate_from_keyfile('''[connection]
id=Test
uuid={}
type={}

[ipv4]
method=manual
address1=192.168.123.123/24
'''.format(UUID, 'dummy'))      # wokeignore:rule=dummy
        self.assert_netplan({UUID: '''network:
  version: 2
  dummy-devices:            # wokeignore:rule=dummy
    NM-{}:
      renderer: NetworkManager
      addresses:
      - "192.168.123.123/24"
      networkmanager:
        uuid: "{}"
        name: "Test"
'''.format(UUID, UUID)})

    def _template_keyfile_type(self, nd_type, nm_type):
        self.maxDiff = None
        extra = ''
        file = os.path.join(self.workdir.name, 'tmp/some.keyfile')
        os.makedirs(os.path.dirname(file))
        with open(file, 'w') as f:
            f.write('[connection]\ntype={}\nuuid={}'.format(nm_type, UUID))
            if nm_type == 'ip-tunnel':
                f.write('\n\n[ip-tunnel]\nmode=4\nremote=10.0.0.1')
                extra = '\n      mode: "isatap"\n      remote: "10.0.0.1"'
        if nd_type in ['ethernets', 'modems', 'wifis']:
            extra = '\n      match: {}'
        parser = netplan.Parser()
        state = netplan.State()
        parser.load_keyfile(file)
        state.import_parser_results(parser)
        output_file = '90-NM-{}.yaml'.format(UUID)
        state._write_yaml_file(output_file, self.workdir.name)
        self.assertTrue(os.path.isfile(os.path.join(self.confdir, output_file)))
        with open(os.path.join(self.confdir, output_file), 'r') as f:
            self.assertEqual(f.read(), '''network:
  version: 2
  {}:
    NM-{}:
      renderer: NetworkManager{}
      networkmanager:
        uuid: "{}"
'''.format(nd_type, UUID, extra, UUID))

    def test_keyfile_ethernet(self):
        self._template_keyfile_type('ethernets', 'ethernet')

    def test_keyfile_type_modem_gsm(self):
        self._template_keyfile_type('modems', 'gsm')

    def test_keyfile_type_modem_cdma(self):
        self._template_keyfile_type('modems', 'cdma')

    def test_keyfile_type_bridge(self):
        self._template_keyfile_type('bridges', 'bridge')

    def test_keyfile_type_bond(self):
        self._template_keyfile_type('bonds', 'bond')

    def test_keyfile_type_tunnel(self):
        self._template_keyfile_type('tunnels', 'ip-tunnel')

    def test_keyfile_type_wifi(self):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
permissions=
id=myid with spaces
interface-name=eth0

[wifi]
ssid=SOME-SSID
mode=infrastructure
hidden=true
mtu=1500
cloned-mac-address=00:11:22:33:44:55
band=a
channel=12
bssid=de:ad:be:ef:ca:fe

[wifi-security]
key-mgmt=ieee8021x

[ipv4]
method=auto
dns-search='''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth0"
      dhcp4: true
      macaddress: "00:11:22:33:44:55"
      mtu: 1500
      access-points:
        "SOME-SSID":
          hidden: true
          bssid: "de:ad:be:ef:ca:fe"
          band: "5GHz"
          channel: 12
          auth:
            key-management: "802.1x"
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"
            passthrough:
              connection.permissions: ""
              ipv4.dns-search: ""
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID, UUID)})

    def _template_keyfile_type_wifi_eap(self, method):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
permissions=
id=testnet
interface-name=wlan0

[wifi]
ssid=testnet
mode=infrastructure

[wifi-security]
key-mgmt=wpa-eap

[802-1x]
eap={}
identity=some-id
anonymous-identity=anon-id
password=v3rys3cr3t!
ca-cert=/some/path.key
client-cert=/some/path.client_cert
private-key=/some/path.key
private-key-password=s0s3cr3t!!111
phase2-auth=chap

[ipv4]
method=auto
dns-search='''.format(UUID, method))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlan0"
      dhcp4: true
      access-points:
        "testnet":
          auth:
            key-management: "eap"
            method: "{}"
            anonymous-identity: "anon-id"
            identity: "some-id"
            ca-certificate: "/some/path.key"
            client-certificate: "/some/path.client_cert"
            client-key: "/some/path.key"
            client-key-password: "s0s3cr3t!!111"
            phase2-auth: "chap"
            password: "v3rys3cr3t!"
          networkmanager:
            uuid: "{}"
            name: "testnet"
            passthrough:
              connection.permissions: ""
              ipv4.dns-search: ""
      networkmanager:
        uuid: "{}"
        name: "testnet"
'''.format(UUID, method, UUID, UUID)})

    def test_keyfile_type_wifi_eap_peap(self):
        self._template_keyfile_type_wifi_eap('peap')

    def test_keyfile_type_wifi_eap_tls(self):
        self._template_keyfile_type_wifi_eap('tls')

    def test_keyfile_type_wifi_eap_ttls(self):
        self._template_keyfile_type_wifi_eap('ttls')

    def test_keyfile_type_wifi_eap_leap(self):
        self._template_keyfile_type_wifi_eap('leap')

    def test_keyfile_type_wifi_eap_pwd(self):
        self._template_keyfile_type_wifi_eap('pwd')

    def test_keyfile_wifi_eap_leap(self):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
permissions=
id=myid with spaces
interface-name=eth0

[wifi]
ssid=SOME-SSID
mode=infrastructure

[wifi-security]
key-mgmt=ieee8021x

[802-1x]
eap=leap
identity=some-id
password=v3rys3cr3t!

[ipv4]
method=auto'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth0"
      dhcp4: true
      access-points:
        "SOME-SSID":
          auth:
            key-management: "802.1x"
            method: "leap"
            identity: "some-id"
            password: "v3rys3cr3t!"
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"
            passthrough:
              connection.permissions: ""
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID, UUID)})

    def test_keyfile_wifi_eap_pwd(self):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
permissions=
id=myid with spaces
interface-name=eth0

[wifi]
ssid=SOME-SSID
mode=infrastructure

[wifi-security]
key-mgmt=ieee8021x

[802-1x]
eap=pwd
identity=some-id
password=v3rys3cr3t!

[ipv4]
method=auto'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth0"
      dhcp4: true
      access-points:
        "SOME-SSID":
          auth:
            key-management: "802.1x"
            method: "pwd"
            identity: "some-id"
            password: "v3rys3cr3t!"
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"
            passthrough:
              connection.permissions: ""
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID, UUID)})

    def test_keyfile_wifi_eap_md5_not_supported(self):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
permissions=
id=myid with spaces
interface-name=eth0

[wifi]
ssid=SOME-SSID
mode=infrastructure

[wifi-security]
key-mgmt=ieee8021x

[802-1x]
eap=md5
identity=some-id
password=v3rys3cr3t!

[ipv4]
method=auto'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth0"
      dhcp4: true
      access-points:
        "SOME-SSID":
          auth:
            key-management: "802.1x"
            identity: "some-id"
            password: "v3rys3cr3t!"
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"
            passthrough:
              connection.permissions: ""
              802-1x.eap: "md5"
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID, UUID)})

    def test_keyfile_wifi_eap_psk_with_eap(self):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
permissions=
id=myid with spaces
interface-name=eth0

[wifi]
ssid=SOME-SSID
mode=infrastructure

[wifi-security]
key-mgmt=wpa-eap
psk=pskpassword

[802-1x]
eap=leap
identity=some-id
password=v3rys3cr3t!

[ipv4]
method=auto'''.format(UUID), regenerate=False)
        # Regeneration is disabled for the previous test because when both,
        # identity and PSK passwords are used, the PSK will be stored in the
        # access-points.<ap>.password key and, because of that, a keyfile with
        # [wifi-security].pmf=2 will be emitted and will differ from the original
        # keyfile.

        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth0"
      dhcp4: true
      access-points:
        "SOME-SSID":
          password: "pskpassword"
          auth:
            key-management: "eap"
            method: "leap"
            identity: "some-id"
            password: "v3rys3cr3t!"
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"
            passthrough:
              connection.permissions: ""
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID, UUID)})

    def _template_keyfile_type_wifi(self, nd_mode, nm_mode):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
id=myid with spaces

[ipv4]
method=auto

[wifi]
ssid=SOME-SSID
wake-on-wlan=24
band=bg
mode={}'''.format(UUID, nm_mode))
        wifi_mode = ''
        ap_mode = ''
        if nm_mode != nd_mode:
            wifi_mode = '''
            passthrough:
              wifi.mode: "{}"'''.format(nm_mode)
        if nd_mode != 'infrastructure':
            ap_mode = '\n          mode: "%s"' % nd_mode
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      dhcp4: true
      wakeonwlan:
      - magic_pkt
      - gtk_rekey_failure
      access-points:
        "SOME-SSID":
          band: "2.4GHz"{}
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"{}
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, ap_mode, UUID, wifi_mode, UUID)})

    def test_keyfile_type_wifi_ap(self):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
id=myid with spaces

[ipv4]
method=shared

[wifi]
ssid=SOME-SSID
wake-on-wlan=24
band=bg
mode=ap'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      wakeonwlan:
      - magic_pkt
      - gtk_rekey_failure
      access-points:
        "SOME-SSID":
          band: "2.4GHz"
          mode: "ap"
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID, UUID)})

    def test_keyfile_type_wifi_adhoc(self):
        self._template_keyfile_type_wifi('adhoc', 'adhoc')

    def test_keyfile_type_wifi_unknown(self):
        self._template_keyfile_type_wifi('infrastructure', 'mesh')

    def test_keyfile_type_wifi_missing_ssid(self):
        err = self.generate_from_keyfile('''[connection]\ntype=wifi\nuuid={}\nid=myid with spaces'''
                                         .format(UUID), expect_fail=True)
        self.assertFalse(os.path.isfile(os.path.join(self.confdir, '90-NM-{}.yaml'.format(UUID))))
        self.assertIn('netplan: Keyfile: cannot find SSID for WiFi connection', err)

    def test_keyfile_wake_on_lan(self):
        self.generate_from_keyfile('''[connection]
type=ethernet
uuid={}
id=myid with spaces

[ethernet]
wake-on-lan=2

[ipv4]
method=auto'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      dhcp4: true
      wakeonlan: true
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
        passthrough:
          ethernet.wake-on-lan: "2"
'''.format(UUID, UUID)})

    def test_keyfile_wake_on_lan_nm_default(self):
        self.generate_from_keyfile('''[connection]
type=ethernet
uuid={}
id=myid with spaces

[ethernet]
wake-on-lan=0

[ipv4]
method=auto'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      dhcp4: true
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID)})

    def test_keyfile_modem_gsm(self):
        self.generate_from_keyfile('''[connection]
type=gsm
uuid={}
id=myid with spaces

[ipv4]
method=auto

[gsm]
auto-config=true'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  modems:
    NM-{}:
      renderer: NetworkManager
      match: {{}}
      dhcp4: true
      auto-config: true
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID)})

    def test_keyfile_existing_id(self):
        self.generate_from_keyfile('''[connection]
type=bridge
interface-name=mybr
uuid={}
id=renamed netplan bridge

[ipv4]
method=auto'''.format(UUID), netdef_id='mybr')
        self.assert_netplan({UUID: '''network:
  version: 2
  bridges:
    mybr:
      renderer: NetworkManager
      dhcp4: true
      networkmanager:
        uuid: "{}"
        name: "renamed netplan bridge"
'''.format(UUID)})

    def test_keyfile_yaml_wifi_hotspot(self):
        self.generate_from_keyfile('''[connection]
id=Hotspot-1
type=wifi
uuid={}
interface-name=wlan0
autoconnect=false
permissions=

[ipv4]
method=shared
dns-search=

[ipv6]
method=ignore
addr-gen-mode=1
dns-search=
ip6-privacy=0

[wifi]
ssid=my-hotspot
mode=ap
mac-address-blacklist= # wokeignore:rule=blacklist

[wifi-security]
group=ccmp;
key-mgmt=wpa-psk
pairwise=ccmp;
proto=rsn;
psk=test1234

[proxy]'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlan0"
      access-points:
        "my-hotspot":
          auth:
            key-management: "psk"
            password: "test1234"
          mode: "ap"
          networkmanager:
            uuid: "{}"
            name: "Hotspot-1"
            passthrough:
              connection.autoconnect: "false"
              connection.permissions: ""
              ipv4.dns-search: ""
              ipv6.addr-gen-mode: "1"
              ipv6.dns-search: ""
              wifi.mac-address-blacklist: "" # wokeignore:rule=blacklist
              wifi-security.group: "ccmp;"
              wifi-security.pairwise: "ccmp;"
              wifi-security.proto: "rsn;"
              proxy._: ""
      networkmanager:
        uuid: "{}"
        name: "Hotspot-1"
'''.format(UUID, UUID, UUID)})

    def test_keyfile_ip4_linklocal_ip6_ignore(self):
        self.generate_from_keyfile('''[connection]
id=netplan-eth1
type=ethernet
interface-name=eth1
uuid={}

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth1"
      networkmanager:
        uuid: "{}"
        name: "netplan-eth1"
'''.format(UUID, UUID)})

    def test_keyfile_vlan(self):
        self.generate_from_keyfile('''[connection]
id=netplan-enblue
type=vlan
interface-name=enblue
uuid={}

[vlan]
id=1
parent=en1

[ipv4]
method=manual
address1=1.2.3.4/24

[ipv6]
method=ignore
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  vlans:
    NM-{}:
      renderer: NetworkManager
      addresses:
      - "1.2.3.4/24"
      id: 1
      link: "en1"
      networkmanager:
        uuid: "{}"
        name: "netplan-enblue"
        passthrough:
          connection.interface-name: "enblue"
'''.format(UUID, UUID)})

    def test_keyfile_bridge(self):
        self.generate_from_keyfile('''[connection]
id=netplan-br0
type=bridge
interface-name=br0
uuid={}

[bridge]
ageing-time=50
priority=1000
forward-delay=12
hello-time=6
max-age=24
stp=false

[ipv4]
method=auto

[ipv6]
method=ignore
'''.format(UUID), netdef_id='br0', expect_fail=False, filename="netplan-br0.nmconnection")
        self.assert_netplan({UUID: '''network:
  version: 2
  bridges:
    br0:
      renderer: NetworkManager
      dhcp4: true
      parameters:
        ageing-time: "50"
        forward-delay: "12"
        hello-time: "6"
        max-age: "24"
        priority: 1000
        stp: false
      networkmanager:
        uuid: "{}"
        name: "netplan-br0"
'''.format(UUID)})

    def test_keyfile_bridge_default_stp(self):
        self.generate_from_keyfile('''[connection]
id=netplan-br0
type=bridge
interface-name=br0
uuid={}

[bridge]
hello-time=6

[ipv4]
method=auto

[ipv6]
method=ignore
'''.format(UUID), netdef_id='br0')
        self.assert_netplan({UUID: '''network:
  version: 2
  bridges:
    br0:
      renderer: NetworkManager
      dhcp4: true
      parameters:
        hello-time: "6"
      networkmanager:
        uuid: "{}"
        name: "netplan-br0"
'''.format(UUID)})

    def test_keyfile_bond(self):
        self.generate_from_keyfile('''[connection]
uuid={}
id=netplan-bn0
type=bond
interface-name=bn0

[bond]
mode=802.3ad
lacp_rate=fast
miimon=10
min_links=10
xmit_hash_policy=none
ad_select=none
all_slaves_active=1 # wokeignore:rule=slave
arp_interval=10
arp_ip_target=10.10.10.10,20.20.20.20
arp_validate=all
arp_all_targets=all
updelay=10
downdelay=10
fail_over_mac=none
num_grat_arp=10
num_unsol_na=10
packets_per_slave=10 # wokeignore:rule=slave
primary_reselect=none
resend_igmp=10
lp_interval=10

[ipv4]
method=auto

[ipv6]
method=ignore
'''.format(UUID), netdef_id='bn0', expect_fail=False, filename='some.keyfile')
        self.assert_netplan({UUID: '''network:
  version: 2
  bonds:
    bn0:
      renderer: NetworkManager
      dhcp4: true
      parameters:
        mode: "802.3ad"
        mii-monitor-interval: "10"
        up-delay: "10"
        down-delay: "10"
        lacp-rate: "fast"
        transmit-hash-policy: "none"
        ad-select: "none"
        arp-validate: "all"
        arp-all-targets: "all"
        fail-over-mac-policy: "none"
        primary-reselect-policy: "none"
        learn-packet-interval: "10"
        arp-interval: "10"
        min-links: 10
        all-members-active: true
        gratuitous-arp: 10
        packets-per-member: 10
        resend-igmp: 10
        arp-ip-targets:
        - 10.10.10.10
        - 20.20.20.20
      networkmanager:
        uuid: "{}"
        name: "netplan-bn0"
'''.format(UUID)})

    def test_keyfile_customer_A1(self):
        self.generate_from_keyfile('''[connection]
id=netplan-wlan0-TESTSSID
type=wifi
interface-name=wlan0
uuid={}

[ipv4]
method=auto

[ipv6]
method=ignore

[wifi]
ssid=TESTSSID
mode=infrastructure

[wifi-security]
key-mgmt=wpa-psk
psk=s0s3cr1t
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlan0"
      dhcp4: true
      access-points:
        "TESTSSID":
          auth:
            key-management: "psk"
            password: "s0s3cr1t"
          networkmanager:
            uuid: "ff9d6ebc-226d-4f82-a485-b7ff83b9607f"
            name: "netplan-wlan0-TESTSSID"
      networkmanager:
        uuid: "{}"
        name: "netplan-wlan0-TESTSSID"
'''.format(UUID, UUID)})

    def test_keyfile_customer_A2(self):
        self.generate_from_keyfile('''[connection]
id=gsm
type=gsm
uuid={}
interface-name=cdc-wdm1

[gsm]
apn=internet

[ipv4]
method=auto
address1=10.10.28.159/24
address2=10.10.164.254/24
address3=10.10.246.132/24
dns=8.8.8.8;8.8.4.4;

[ipv6]
method=auto
addr-gen-mode=1
ip6-privacy=0
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  modems:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "cdc-wdm1"
      nameservers:
        addresses:
        - 8.8.8.8
        - 8.8.4.4
      dhcp4: true
      dhcp6: true
      apn: "internet"
      networkmanager:
        uuid: "{}"
        name: "gsm"
        passthrough:
          ipv4.address1: "10.10.28.159/24"
          ipv4.address2: "10.10.164.254/24"
          ipv4.address3: "10.10.246.132/24"
          ipv6.addr-gen-mode: "1"
'''.format(UUID, UUID)})

    def test_keyfile_netplan0103_compat(self):
        self.generate_from_keyfile('''[connection]
id=Work Wired
uuid={}
type=ethernet
autoconnect=false
permissions=
timestamp=305419896

[ethernet]
mac-address=99:88:77:66:55:44
mac-address-blacklist= # wokeignore:rule=blacklist
mtu=900

[ipv4]
address1=192.168.0.5/24,192.168.0.1
address2=1.2.3.4/8
dns=4.2.2.1;4.2.2.2;
dns-search=
method=manual
route1=10.10.10.2/24,10.10.10.1,3
route2=1.1.1.1/8,1.2.1.1,1
route3=2.2.2.2/7
route4=3.3.3.3/6,0.0.0.0,4
route4_options=cwnd=10,mtu=1492,src=1.2.3.4

[ipv6]
addr-gen-mode=stable-privacy
address1=abcd::beef/64
address2=dcba::beef/56
dns=1::cafe;2::cafe;
dns-search=wallaceandgromit.com;
method=manual
ip6-privacy=1
route1=1:2:3:4:5:6:7:8/64,8:7:6:5:4:3:2:1,3
route2=2001::1000/56,2001::1111,1
route3=4:5:6:7:8:9:0:1/63,::,5
route4=5:6:7:8:9:0:1:2/62

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        macaddress: "99:88:77:66:55:44"
      addresses:
      - "192.168.0.5/24"
      - "1.2.3.4/8"
      - "abcd::beef/64"
      - "dcba::beef/56"
      nameservers:
        addresses:
        - 4.2.2.1
        - 4.2.2.2
        - 1::cafe
        - 2::cafe
      ipv6-address-generation: "stable-privacy"
      mtu: 900
      routes:
      - metric: 3
        to: "10.10.10.2/24"
        via: "10.10.10.1"
      - metric: 1
        to: "1.1.1.1/8"
        via: "1.2.1.1"
      - scope: "link"
        to: "2.2.2.2/7"
      - scope: "link"
        metric: 4
        mtu: 1492
        from: "1.2.3.4"
        to: "3.3.3.3/6"
      - metric: 3
        to: "1:2:3:4:5:6:7:8/64"
        via: "8:7:6:5:4:3:2:1"
      - metric: 1
        to: "2001::1000/56"
        via: "2001::1111"
      - scope: "link"
        metric: 5
        to: "4:5:6:7:8:9:0:1/63"
      - scope: "link"
        to: "5:6:7:8:9:0:1:2/62"
      wakeonlan: true
      networkmanager:
        uuid: "{}"
        name: "Work Wired"
        passthrough:
          connection.autoconnect: "false"
          connection.permissions: ""
          connection.timestamp: "305419896"
          ethernet.mac-address-blacklist: "" # wokeignore:rule=blacklist
          ipv4.address1: "192.168.0.5/24,192.168.0.1"
          ipv4.dns-search: ""
          ipv4.method: "manual"
          ipv4.route4: "3.3.3.3/6,0.0.0.0,4"
          ipv4.route4_options: "cwnd=10,mtu=1492,src=1.2.3.4"
          ipv6.dns-search: "wallaceandgromit.com;"
          ipv6.ip6-privacy: "1"
          proxy._: ""
'''.format(UUID, UUID)})

    def test_keyfile_tunnel_regression_lp1952967(self):
        self.generate_from_keyfile('''[connection]
id=IP tunnel connection 1
uuid={}
type=ip-tunnel
autoconnect=false
interface-name=gre10
permissions=

[ip-tunnel]
local=10.20.20.1
mode=2
remote=10.20.20.2

[ipv4]
dns-search=
method=auto

[ipv6]
addr-gen-mode=stable-privacy
dns-search=
method=auto

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      dhcp6: true
      ipv6-address-generation: "stable-privacy"
      mode: "gre"
      local: "10.20.20.1"
      remote: "10.20.20.2"
      networkmanager:
        uuid: "{}"
        name: "IP tunnel connection 1"
        passthrough:
          connection.autoconnect: "false"
          connection.interface-name: "gre10"
          connection.permissions: ""
          ipv4.dns-search: ""
          ipv6.dns-search: ""
          ipv6.ip6-privacy: "-1"
          proxy._: ""
'''.format(UUID, UUID)})

    def test_keyfile_ip6_privacy_default_netplan_0104_compat(self):
        self.generate_from_keyfile('''[connection]
id=Test
uuid={}
type=ethernet

[ethernet]
mac-address=99:88:77:66:55:44

[ipv4]
method=auto

[ipv6]
method=auto
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        macaddress: "99:88:77:66:55:44"
      dhcp4: true
      dhcp6: true
      wakeonlan: true
      networkmanager:
        uuid: "{}"
        name: "Test"
        passthrough:
          ipv6.ip6-privacy: "-1"
'''.format(UUID, UUID)})

    def test_keyfile_wpa3_sae(self):
        self.generate_from_keyfile('''[connection]
id=test2
uuid={}
type=wifi
interface-name=wlan0

[wifi]
mode=infrastructure
ssid=ubuntu-wpa2-wpa3-mixed

[wifi-security]
key-mgmt=sae
psk=test1234
pmf=3

[ipv4]
method=auto

[ipv6]
addr-gen-mode=stable-privacy
method=auto

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlan0"
      dhcp4: true
      dhcp6: true
      ipv6-address-generation: "stable-privacy"
      access-points:
        "ubuntu-wpa2-wpa3-mixed":
          auth:
            key-management: "sae"
            password: "test1234"
          networkmanager:
            uuid: "ff9d6ebc-226d-4f82-a485-b7ff83b9607f"
            name: "test2"
            passthrough:
              ipv6.ip6-privacy: "-1"
              proxy._: ""
      networkmanager:
        uuid: "{}"
        name: "test2"
'''.format(UUID, UUID)})

    def test_keyfile_wpa_sha256(self):
        self.generate_from_keyfile('''[connection]
id=test2
uuid={}
type=wifi
interface-name=wlan0

[wifi]
mode=infrastructure
ssid=ubuntu-wpa-sha256

[wifi-security]
key-mgmt=wpa-psk
pmf=2
psk=test1234

[ipv4]
method=auto

[ipv6]
addr-gen-mode=stable-privacy
method=auto

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlan0"
      dhcp4: true
      dhcp6: true
      ipv6-address-generation: "stable-privacy"
      access-points:
        "ubuntu-wpa-sha256":
          auth:
            key-management: "psk-sha256"
            password: "test1234"
          networkmanager:
            uuid: "ff9d6ebc-226d-4f82-a485-b7ff83b9607f"
            name: "test2"
            passthrough:
              ipv6.ip6-privacy: "-1"
              proxy._: ""
      networkmanager:
        uuid: "{}"
        name: "test2"
'''.format(UUID, UUID)})

    def test_keyfile_wpa3_enterprise_eap_sha256(self):
        self.generate_from_keyfile('''[connection]
id=test2
uuid={}
type=wifi
interface-name=wlan0

[wifi]
mode=infrastructure
ssid=enterprisenet

[wifi-security]
key-mgmt=wpa-eap
pmf=2

[802-1x]
eap=tls
identity=cert-joe@cust.example.com
anonymous-identity=@cust.example.com
ca-cert=/etc/ssl/cust-cacrt.pem
client-cert=/etc/ssl/cust-crt.pem
private-key=/etc/ssl/cust-key.pem
private-key-password=**********

[ipv4]
method=auto

[ipv6]
addr-gen-mode=stable-privacy
method=auto

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlan0"
      dhcp4: true
      dhcp6: true
      ipv6-address-generation: "stable-privacy"
      access-points:
        "enterprisenet":
          auth:
            key-management: "eap-sha256"
            method: "tls"
            anonymous-identity: "@cust.example.com"
            identity: "cert-joe@cust.example.com"
            ca-certificate: "/etc/ssl/cust-cacrt.pem"
            client-certificate: "/etc/ssl/cust-crt.pem"
            client-key: "/etc/ssl/cust-key.pem"
            client-key-password: "**********"
          networkmanager:
            uuid: "ff9d6ebc-226d-4f82-a485-b7ff83b9607f"
            name: "test2"
            passthrough:
              ipv6.ip6-privacy: "-1"
              proxy._: ""
      networkmanager:
        uuid: "{}"
        name: "test2"
'''.format(UUID, UUID)})

    def test_keyfile_wpa3_enterprise_eap_suite_b_192(self):
        self.generate_from_keyfile('''[connection]
id=test2
uuid={}
type=wifi
interface-name=wlan0

[wifi]
mode=infrastructure
ssid=enterprisenet

[wifi-security]
key-mgmt=wpa-eap-suite-b-192
pmf=3

[802-1x]
eap=tls
identity=cert-joe@cust.example.com
anonymous-identity=@cust.example.com
ca-cert=/etc/ssl/cust-cacrt.pem
client-cert=/etc/ssl/cust-crt.pem
private-key=/etc/ssl/cust-key.pem
private-key-password=**********

[ipv4]
method=auto

[ipv6]
addr-gen-mode=stable-privacy
method=auto

[proxy]
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlan0"
      dhcp4: true
      dhcp6: true
      ipv6-address-generation: "stable-privacy"
      access-points:
        "enterprisenet":
          auth:
            key-management: "eap-suite-b-192"
            method: "tls"
            anonymous-identity: "@cust.example.com"
            identity: "cert-joe@cust.example.com"
            ca-certificate: "/etc/ssl/cust-cacrt.pem"
            client-certificate: "/etc/ssl/cust-crt.pem"
            client-key: "/etc/ssl/cust-key.pem"
            client-key-password: "**********"
          networkmanager:
            uuid: "ff9d6ebc-226d-4f82-a485-b7ff83b9607f"
            name: "test2"
            passthrough:
              ipv6.ip6-privacy: "-1"
              proxy._: ""
      networkmanager:
        uuid: "{}"
        name: "test2"
'''.format(UUID, UUID)})

    def test_keyfile_dns_search_ip4_ip6_conflict(self):
        self.generate_from_keyfile('''[connection]
id=Work Wired
type=ethernet
uuid={}
autoconnect=false
timestamp=305419896

[ethernet]
wake-on-lan=1
mac-address=99:88:77:66:55:44
mtu=900

[ipv4]
method=manual
address1=192.168.0.5/24,192.168.0.1
address2=1.2.3.4/8
dns=4.2.2.1;4.2.2.2;

[ipv6]
method=manual
address1=abcd::beef/64
address2=dcba::beef/56
addr-gen-mode=1
dns=1::cafe;2::cafe;
dns-search=wallaceandgromit.com;

[proxy]\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        macaddress: "99:88:77:66:55:44"
      addresses:
      - "192.168.0.5/24"
      - "1.2.3.4/8"
      - "abcd::beef/64"
      - "dcba::beef/56"
      nameservers:
        addresses:
        - 4.2.2.1
        - 4.2.2.2
        - 1::cafe
        - 2::cafe
      mtu: 900
      wakeonlan: true
      networkmanager:
        uuid: "{}"
        name: "Work Wired"
        passthrough:
          connection.autoconnect: "false"
          connection.timestamp: "305419896"
          ethernet.wake-on-lan: "1"
          ipv4.method: "manual"
          ipv4.address1: "192.168.0.5/24,192.168.0.1"
          ipv6.addr-gen-mode: "1"
          ipv6.dns-search: "wallaceandgromit.com;"
          ipv6.ip6-privacy: "-1"
          proxy._: ""
'''.format(UUID, UUID)})

    def test_keyfile_nm_140_default_ethernet_group(self):
        self.generate_from_keyfile('''[connection]
id=Test Write Bridge Main
uuid={}
type=bridge
interface-name=br0

[ethernet]

[bridge]

[ipv4]
address1=1.2.3.4/24,1.1.1.1
method=manual

[ipv6]
addr-gen-mode=default
method=auto

[proxy]\n'''.format(UUID), netdef_id='br0')
        self.assert_netplan({UUID: '''network:
  version: 2
  bridges:
    br0:
      renderer: NetworkManager
      addresses:
      - "1.2.3.4/24"
      dhcp6: true
      networkmanager:
        uuid: "{}"
        name: "Test Write Bridge Main"
        passthrough:
          ethernet._: ""
          bridge._: ""
          ipv4.address1: "1.2.3.4/24,1.1.1.1"
          ipv4.method: "manual"
          ipv6.addr-gen-mode: "default"
          ipv6.ip6-privacy: "-1"
          proxy._: ""
'''.format(UUID)})

    def test_multiple_eap_methods(self):
        self.generate_from_keyfile('''[connection]
id=MyWifi
uuid={}
type=wifi
interface-name=wlp2s0

[wifi]
mode=infrastructure
ssid=MyWifi

[wifi-security]
auth-alg=open
key-mgmt=wpa-eap

[802-1x]
ca-cert=/path/to/my/crt.crt
eap=peap;tls
identity=username
password=123456
phase2-auth=mschapv2

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlp2s0"
      dhcp4: true
      access-points:
        "MyWifi":
          auth:
            key-management: "eap"
            method: "peap"
            identity: "username"
            ca-certificate: "/path/to/my/crt.crt"
            phase2-auth: "mschapv2"
            password: "123456"
          networkmanager:
            uuid: "{}"
            name: "MyWifi"
            passthrough:
              wifi-security.auth-alg: "open"
              802-1x.eap: "peap;tls"
      networkmanager:
        uuid: "{}"
        name: "MyWifi"
'''.format(UUID, UUID, UUID)})

    def test_single_eap_method(self):
        self.generate_from_keyfile('''[connection]
id=MyWifi
uuid={}
type=wifi
interface-name=wlp2s0

[wifi]
mode=infrastructure
ssid=MyWifi

[wifi-security]
auth-alg=open
key-mgmt=wpa-eap

[802-1x]
ca-cert=/path/to/my/crt.crt
eap=peap;
identity=username
password=123456
phase2-auth=mschapv2

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "wlp2s0"
      dhcp4: true
      access-points:
        "MyWifi":
          auth:
            key-management: "eap"
            method: "peap"
            identity: "username"
            ca-certificate: "/path/to/my/crt.crt"
            phase2-auth: "mschapv2"
            password: "123456"
          networkmanager:
            uuid: "{}"
            name: "MyWifi"
            passthrough:
              wifi-security.auth-alg: "open"
      networkmanager:
        uuid: "{}"
        name: "MyWifi"
'''.format(UUID, UUID, UUID)})

    def test_simple_wireguard(self):
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_wireguard_with_key(self):
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[wireguard]
private-key=aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A=

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      keys:
        private: "aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A="
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_wireguard_with_key_and_peer(self):
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[wireguard]
private-key=aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A=

[wireguard-peer.cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI=]
endpoint=1.2.3.4:12345
allowed-ips=192.168.0.0/24;

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      keys:
        private: "aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A="
      peers:
      - endpoint: "1.2.3.4:12345"
        keys:
          public: "cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI="
        allowed-ips:
        - "192.168.0.0/24"
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_wireguard_with_empty_endpoint(self):
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[wireguard]
private-key=aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A=

[wireguard-peer.cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI=]
endpoint=
allowed-ips=192.168.0.0/24;

[ipv4]
method=auto\n'''.format(UUID), regenerate=False)
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      keys:
        private: "aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A="
      peers:
      - keys:
          public: "cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI="
        allowed-ips:
        - "192.168.0.0/24"
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_wireguard_allowed_ips_without_prefix(self):
        '''
        When the IP prefix is not present we should default to /32 for IPv4
        and /128 for IPv6.
        '''
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[wireguard]
private-key=aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A=

[wireguard-peer.cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI=]
endpoint=1.2.3.4:12345
allowed-ips=192.168.0.10;2001::1;

[ipv4]
method=auto\n'''.format(UUID), regenerate=False)
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      keys:
        private: "aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A="
      peers:
      - endpoint: "1.2.3.4:12345"
        keys:
          public: "cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI="
        allowed-ips:
        - "192.168.0.10/32"
        - "2001::1/128"
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_wireguard_with_key_flags(self):
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[wireguard]
listen-port=51820
private-key-flags=1

[wireguard-peer.M9nt4YujIOmNrRmpIRTmYSfMdrpvE7u6WkG8FY8WjG4=]
endpoint=10.20.30.40:51820
allowed-ips=0.0.0.0/0;

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      port: 51820
      keys:
        private-key-flags:
        - agent-owned
      peers:
      - endpoint: "10.20.30.40:51820"
        keys:
          public: "M9nt4YujIOmNrRmpIRTmYSfMdrpvE7u6WkG8FY8WjG4="
        allowed-ips:
        - "0.0.0.0/0"
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_wireguard_with_key_all_flags_enabled(self):
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[wireguard]
listen-port=51820
private-key-flags=7

[wireguard-peer.M9nt4YujIOmNrRmpIRTmYSfMdrpvE7u6WkG8FY8WjG4=]
endpoint=10.20.30.40:51820
allowed-ips=0.0.0.0/0;

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      port: 51820
      keys:
        private-key-flags:
        - agent-owned
        - not-saved
        - not-required
      peers:
      - endpoint: "10.20.30.40:51820"
        keys:
          public: "M9nt4YujIOmNrRmpIRTmYSfMdrpvE7u6WkG8FY8WjG4="
        allowed-ips:
        - "0.0.0.0/0"
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_wireguard_with_key_and_peer_without_allowed_ips(self):
        self.generate_from_keyfile('''[connection]
id=wg0
type=wireguard
uuid={}
interface-name=wg0

[wireguard]
private-key=aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A=

[wireguard-peer.cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI=]
endpoint=1.2.3.4:12345

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "wireguard"
      keys:
        private: "aPUcp5vHz8yMLrzk8SsDyYnV33IhE/k20e52iKJFV0A="
      peers:
      - endpoint: "1.2.3.4:12345"
        keys:
          public: "cwkb7k0xDgLSnunZpFIjLJw4u+mJDDr+aBR5DqzpmgI="
      networkmanager:
        uuid: "{}"
        name: "wg0"
        passthrough:
          connection.interface-name: "wg0"
'''.format(UUID, UUID)})

    def test_vxlan_with_local_and_remote(self):
        self.generate_from_keyfile('''[connection]
id=vxlan10
type=vxlan
uuid={}
interface-name=vxlan10

[vxlan]
id=10
local=198.51.100.2
remote=203.0.113.1

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "vxlan"
      local: "198.51.100.2"
      remote: "203.0.113.1"
      id: 10
      networkmanager:
        uuid: "{}"
        name: "vxlan10"
        passthrough:
          connection.interface-name: "vxlan10"
'''.format(UUID, UUID)})

    def test_simple_vxlan(self):
        self.generate_from_keyfile('''[connection]
id=vxlan10
type=vxlan
uuid={}
interface-name=vxlan10

[vxlan]
id=10

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  tunnels:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      mode: "vxlan"
      id: 10
      networkmanager:
        uuid: "{}"
        name: "vxlan10"
        passthrough:
          connection.interface-name: "vxlan10"
'''.format(UUID, UUID)})

    def test_invalid_tunnel_mode(self):
        out = self.generate_from_keyfile('''[connection]
id=tun0
type=ip-tunnel
uuid={}
interface-name=tun0

[ip-tunnel]
mode=42

[ipv4]
method=auto\n'''.format(UUID), expect_fail=True)

        self.assertIn('missing or invalid \'mode\' property for tunnel', out)

    def test_keyfile_wifi_random_cloned_mac_address(self):
        self.generate_from_keyfile('''[connection]
type=wifi
uuid={}
id=myid with spaces
interface-name=eth0

[wifi]
ssid=SOME-SSID
mode=infrastructure
cloned-mac-address=random

[wifi-security]
key-mgmt=ieee8021x

[ipv4]
method=auto
dns-search='''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  wifis:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth0"
      dhcp4: true
      macaddress: "random"
      access-points:
        "SOME-SSID":
          auth:
            key-management: "802.1x"
          networkmanager:
            uuid: "{}"
            name: "myid with spaces"
            passthrough:
              ipv4.dns-search: ""
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
'''.format(UUID, UUID, UUID)})

    def test_keyfile_ethernet_random_cloned_mac_address(self):
        self.generate_from_keyfile('''[connection]
type=ethernet
uuid={}
id=myid with spaces
interface-name=eth0

[ethernet]
cloned-mac-address=random

[ipv4]
method=auto
dns-search='''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth0"
      dhcp4: true
      macaddress: "random"
      wakeonlan: true
      networkmanager:
        uuid: "{}"
        name: "myid with spaces"
        passthrough:
          ipv4.dns-search: ""
'''.format(UUID, UUID)})

    def test_veth_pair(self):
        self.generate_from_keyfile('''[connection]
id=veth-peer1
uuid={}
type=veth
interface-name=veth-peer1

[veth]
peer=veth-peer2

[ipv4]
method=auto\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  virtual-ethernets:
    NM-{}:
      renderer: NetworkManager
      dhcp4: true
      peer: "veth-peer2"
      networkmanager:
        uuid: "{}"
        name: "veth-peer1"
        passthrough:
          connection.interface-name: "veth-peer1"
'''.format(UUID, UUID)})

    def test_veth_without_peer(self):
        self.generate_from_keyfile('''[connection]
id=veth-peer1
uuid={}
type=veth
interface-name=veth-peer1

[ipv4]
method=auto\n'''.format(UUID), expect_fail=True)

    def test_vrf_basic(self):
        self.generate_from_keyfile('''[connection]
id=vrf0
uuid={}
type=vrf
interface-name=vrf0

[vrf]
table=1000

[ipv4]
route1=10.10.0.0/16,10.10.10.1
route1_options=table=1000
method=link-local\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  vrfs:
    NM-{}:
      renderer: NetworkManager
      routes:
      - to: "10.10.0.0/16"
        via: "10.10.10.1"
      table: 1000
      networkmanager:
        uuid: "{}"
        name: "vrf0"
        passthrough:
          connection.interface-name: "vrf0"
'''.format(UUID, UUID)})

    def test_vrf_without_table_should_fail(self):
        out = self.generate_from_keyfile('''[connection]
id=vrf0
uuid={}
type=vrf
interface-name=vrf0

[vrf]

[ipv4]
route1=10.10.0.0/16,10.10.10.1
route1_options=table=1000
method=link-local\n'''.format(UUID), expect_fail=True)

        self.assertIn('missing \'table\' property', out)

    def test_nm_parse_advmss_method_manual(self):
        self.generate_from_keyfile('''[connection]
id=netplan-engreen
type=ethernet
uuid={}
interface-name=engreen

[ethernet]
wake-on-lan=0


[ipv4]
method=manual
address1=192.168.14.2/24
route1=10.10.10.0/24,192.168.1.20
route1_options=advmss=1400
'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "engreen"
      addresses:
      - "192.168.14.2/24"
      routes:
      - to: "10.10.10.0/24"
        via: "192.168.1.20"
        advertised-mss: 1400
      networkmanager:
        uuid: "{}"
        name: "netplan-engreen"\n'''.format(UUID, UUID)})

    def test_nameserver_with_DoT_lp2055148(self):
        self.generate_from_keyfile('''[connection]
id=ethernet-eth123
uuid={}
type=ethernet
interface-name=eth123

[ethernet]

[ipv4]
dns=8.8.8.8;1.1.1.1#lxd;192.168.0.1#domain.local;
method=auto

[ipv6]
addr-gen-mode=default
method=auto

[proxy]\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  ethernets:
    NM-{}:
      renderer: NetworkManager
      match:
        name: "eth123"
      dhcp4: true
      dhcp6: true
      wakeonlan: true
      networkmanager:
        uuid: "{}"
        name: "ethernet-eth123"
        passthrough:
          ethernet._: ""
          ipv4.dns: "8.8.8.8;1.1.1.1#lxd;192.168.0.1#domain.local;"
          ipv6.addr-gen-mode: "default"
          ipv6.ip6-privacy: "-1"
          proxy._: ""
'''.format(UUID, UUID)})

    def test_ipv4_route_metric_is_overriden_when_dhcp4_is_disabled(self):
        self.generate_from_keyfile('''[connection]
id=dummy-123
type=dummy
uuid={}
interface-name=dummy123

[ipv4]
method=manual
address1=100.85.0.1/24,100.85.0.1
route-metric=95

[ipv6]
method=ignore
addr-gen-mode=default

[dummy]

[proxy]\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  dummy-devices:
    NM-{}:
      renderer: NetworkManager
      addresses:
      - "100.85.0.1/24"
      dhcp4-overrides:
        route-metric: 95
      networkmanager:
        uuid: "{}"
        name: "dummy-123"
        passthrough:
          connection.interface-name: "dummy123"
          ipv4.method: "manual"
          ipv4.address1: "100.85.0.1/24,100.85.0.1"
          ipv6.addr-gen-mode: "default"
          dummy._: ""
          proxy._: ""
'''.format(UUID, UUID)})

    def test_ipv6_route_metric_is_overriden_when_dhcp6_is_disabled(self):
        self.generate_from_keyfile('''[connection]
id=dummy-123
type=dummy
uuid={}
interface-name=dummy123

[ipv4]
method=disabled

[ipv6]
method=manual
address1=fdeb:446c:912d:8da::/64,fdeb:446c:912d:8da::1
route-metric=95
addr-gen-mode=default

[dummy]

[proxy]\n'''.format(UUID))
        self.assert_netplan({UUID: '''network:
  version: 2
  dummy-devices:
    NM-{}:
      renderer: NetworkManager
      addresses:
      - "fdeb:446c:912d:8da::/64"
      dhcp6-overrides:
        route-metric: 95
      networkmanager:
        uuid: "{}"
        name: "dummy-123"
        passthrough:
          connection.interface-name: "dummy123"
          ipv4.method: "disabled"
          ipv6.method: "manual"
          ipv6.address1: "fdeb:446c:912d:8da::/64,fdeb:446c:912d:8da::1"
          ipv6.addr-gen-mode: "default"
          ipv6.ip6-privacy: "-1"
          dummy._: ""
          proxy._: ""
'''.format(UUID, UUID)})



#!/usr/bin/python3
#
# Copyright (C) 2022 Canonical, Ltd.
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''netplan SR-IOV rebind command line'''

import logging
import os
import sys
from time import sleep

from .. import utils
from ..sriov import PCIDevice, bind_vfs, _get_pci_slot_name
import netplan


FALLBACK_WAIT_TIME_SEC = 3
INTERVAL_SEC = 0.2
MAX_WAITING_TIME_SEC = 5


class MLX5VFLAGStateNotFound(Exception):
    pass


class MLX5VFLAGStateCannotBeRead(Exception):
    pass


class MLX5VFLAGStateDisabled(Exception):
    pass


class NetplanSriovRebind(utils.NetplanCommand):

    def __init__(self):
        super().__init__(command_id='rebind',
                         description='Rebind SR-IOV virtual functions of given physical functions to their driver',
                         leaf=True)

    def run(self):
        self.parser.add_argument('--root-dir', default='/',
                                 help='Search for configuration files in this root directory instead of /')
        self.parser.add_argument('netdevs', type=str, nargs='*', default=[],
                                 help='Space separated list of PF interface names')
        self.func = self.command_rebind

        self.logger = logging.getLogger('sriov_rebind')
        self.logger.propagate = False
        log_handler = logging.StreamHandler(stream=sys.stdout)

        self.parse_args()

        # netplan rebind --debug setup
        if self.debug:
            self.logger.setLevel(logging.DEBUG)
            log_handler.setLevel(logging.DEBUG)
            log_handler.setFormatter(logging.Formatter('%(levelname)s:%(message)s'))
        else:
            self.logger.setLevel(logging.INFO)
            log_handler.setLevel(logging.INFO)
            log_handler.setFormatter(logging.Formatter('%(message)s'))

        self.logger.addHandler(log_handler)

        self.run_command()

    def command_rebind(self):
        """Bind virtual functions of SR-IOV devices to their corresponding driver after eswitch mode was changed"""
        for iface in self.netdevs:
            pci_addr = _get_pci_slot_name(iface)
            pcidev = PCIDevice(pci_addr)
            if not pcidev.is_pf:
                self.logger.debug('{} does not seem to be a SR-IOV physical function'.format(iface))
                continue

            # There are some hardware-specific configuration that must happen *before* the bind
            # of VFs to their drivers. Some settings take time to be effective and, when possible,
            # we need to wait until the driver reports it's ready.
            self._perform_hardware_specific_quirks(iface, pcidev)

            bound_vfs = bind_vfs(pcidev.vfs, pcidev.driver)
            self.logger.debug('{}: bound {} VFs'.format(pcidev, len(bound_vfs)))

    def _perform_hardware_specific_quirks(self, iface: str, pf: PCIDevice):
        """
        Perform any hardware-specific quirks for the given SR-IOV device to make
        sure it's ready before the bind.
        """

        if pf.driver in ['mlx5_core']:
            # Mellanox specific quirks

            parser = netplan.Parser()
            parser.load_yaml_hierarchy(self.root_dir)
            np_state = netplan.State()
            np_state.import_parser_results(parser)

            for netdef in np_state.ethernets.values():
                if (netdef._has_match and netdef.set_name == iface) or netdef.id == iface:
                    if bond_link := netdef.links.get('bond'):
                        # VF LAG support. See LP: #1988018
                        # If the PF is a member of a bond, the user might be trying to enable the
                        # VF LAG feature.
                        # Mellanox VF LAG requires that the LAG state reports as 'active'
                        # *before* VFs can be bound to the driver. Performing the bind operation
                        # before the device is ready will cause the VF LAG feature to never be enabled.

                        # Another condition for the VF LAG activation is that the LAG mode
                        # must be one of 'active-backup', 'balanced-xor' or '802.3ad'.
                        bond_mode = bond_link._bond_mode
                        if not self._is_bond_mode_supported(bond_mode):
                            self.logger.debug(f'{iface} - LAG mode {bond_mode} is not supported by VF LAG')
                            continue

                        self.logger.debug(f'{iface} - waiting for the LAG state to be \'active\'')
                        try:
                            self._wait_for_mlx5_pf_lag_state_active(pf)
                        except MLX5VFLAGStateCannotBeRead:
                            self.logger.debug(f'{iface} - VF LAG state cannot be read')
                        except MLX5VFLAGStateNotFound:
                            self.logger.debug(f'{iface} - VF LAG state debugfs file not found')
                        except MLX5VFLAGStateDisabled:
                            self.logger.debug(f'{iface} - VF LAG state is still \'disabled\' after waiting')
                        else:
                            self.logger.debug(f'{iface} - VF LAG state is \'active\'')

    def _wait_for_mlx5_pf_lag_state_active(self, pf: PCIDevice):
        """
        The mlx5 driver added support for debugfs in https://github.com/torvalds/linux/commit/7f46a0b7327a
        It's available since kernel 5.19 https://cdn.kernel.org/pub/linux/kernel/v5.x/ChangeLog-5.19
        """
        retries = int(MAX_WAITING_TIME_SEC / INTERVAL_SEC)
        pci_addr = pf.pci_addr
        path = f'/sys/kernel/debug/mlx5/{pci_addr}/lag/state'

        if not os.path.exists(path):
            # If the debugfs file doesn't exist, it might be because this version of the mlx5 driver
            # still doesn't support it or because the debugfs is not mounted.
            # In this case, we probably should still wait for a few seconds to give time for the
            # driver to change state.
            # Based on tests with a ConnectX-5 NIC, 1 second is enough time, so let's wait a bit more
            # just in case. This delay will only be introduced if the PF is part of a bond.
            sleep(FALLBACK_WAIT_TIME_SEC)
            raise MLX5VFLAGStateNotFound

        while retries > 0:
            try:
                if self._get_mlx5_vf_lag_state(pci_addr) != 'active':
                    self.logger.debug(f'{pci_addr} VF LAG state is not active yet, retrying...')
                    # Based on tests with a ConnectX-5 NIC, a single 1-second cycle was enough time to
                    # allow the interfaces to change state.
                    sleep(INTERVAL_SEC)
                else:
                    return

            except Exception:
                raise MLX5VFLAGStateCannotBeRead

            retries = retries - 1

        raise MLX5VFLAGStateDisabled

    def _is_bond_mode_supported(self, mode: str) -> bool:
        '''
        Return True or False if the bond mode is one of the supported modes
        for the VG LAG activation.
        '''
        return mode in ['active-backup', 'balanced-xor', '802.3ad']

    def _get_mlx5_vf_lag_state(self, pci_addr: str) -> str:
        path = f'/sys/kernel/debug/mlx5/{pci_addr}/lag/state'

        with open(path, 'r') as f:
            return f.read().strip()
